import rclpy
from rclpy.action import ActionClient

from control_msgs.action import FollowJointTrajectory


class TrajectoryActionClient:
    """A simple trajectory action client."""

    def __init__(
        self,
        parent_node: "rclpy.node.Node",
        action_name: str = "position_joint_trajectory_controller/follow_joint_trajectory",
    ):
        self.node = parent_node
        self._action_client = ActionClient(
            self.node,
            FollowJointTrajectory,
            action_name,
        )

    def send_trajectory_goal(self, goal_msg: FollowJointTrajectory.Goal):
        self._action_client.wait_for_server()

        self._action_client.send_goal_async(goal_msg)



import numpy as np


def normalised(x: np.ndarray) -> np.ndarray:
    """Return a normalised vector (unit vector)

    Args:
        x (np.ndarray): array-like shape-(N,) vector
    Returns:
        (np.ndarray): a unit/normalised vector shape-(N,)

    """
    norm = np.linalg.norm(x)

    if np.isclose(norm, 0.0):
        return x

    return x / norm


def as_transform(
    u: np.ndarray, v: np.ndarray, n: np.ndarray, p: np.ndarray = np.zeros(3)
) -> np.ndarray:
    """Return a 4x4 homogenous transformation matrix constructed from its column vector parts

    Args:
        u (np.ndarray): array-like shape-(3,) vector
        v (np.ndarray): array-like shape-(3,) vector
        n (np.ndarray): array-like shape-(3,) vector
        p (np.ndarray): array-like shape-(3,) vector
    Returns:
        (np.ndarray): An array shape-(4,4) representing a rigid transformation matrix

    """
    frame = np.vstack(
        [
            np.hstack(
                [u.reshape(3, 1), v.reshape(3, 1), n.reshape(3, 1), p.reshape(3, 1)]
            ),
            [0, 0, 0, 1],
        ]
    )

    return frame


def proj(u: np.ndarray, v: np.ndarray) -> np.ndarray:
    """
    Project u over v

    Args:
        u (np.ndarray): array-like shape-(N,) vector
        v (np.ndarray): array-like shape-(N,) vector
    Returns:
        (np.ndarray): An array shape-(N,4) representing projection of u over v
    """

    dot_uv = np.dot(u, v)
    return dot_uv * v / np.dot(v, v)


def plane_vector_directors(plane_normal: np.ndarray) -> np.ndarray:
    """
    Computes the vector directors from a hyper-plane normal vector

    Args:
        plane_normal (np.ndarray): array-like shape-(N,) hyper-plane normal vector
    Returns:
        (np.ndarray): An array shape-(N,N) where each row is a vector director of the hyper-plane
    """

    plane_mat = np.vstack([plane_normal, np.zeros(3), np.zeros(3)])
    U, _, _ = np.linalg.svd(plane_mat, full_matrices=True)

    U[:, 0] = np.zeros(3)

    return U


def project_on_plane(
    plane_origin: np.ndarray, plane_normal: np.ndarray, points: np.ndarray
) -> np.ndarray:
    """
    Projects points on plane

    Args:
        plane_origin (np.ndarray): array-like shape-(N,) hyper-plane origin
        plane_normal (np.ndarray): array-like shape-(N,) hyper-plane normal vector
        points (np.ndarray): array-like shape-(M,N) representing a list of N-dimensional points
    Returns:
        (np.ndarray): array-like shape-(M,N) representing a list of N-dimensional points projected on the plane
    """

    mat = plane_vector_directors(plane_normal)
    flat_points = plane_origin + np.dot(np.dot(points - plane_origin, mat), mat.T)

    return flat_points


def multivariate_normal(
    mean: np.ndarray,
    cov: np.ndarray,
    decomposition_type: str = "cholesky",
    check_valid=False,
    samples=1,
    eps=0.000000001,
):
    """Generates multivariate gaussian noise by sampling the cov matrix and centering
    the distribution about the given mean

    Args:
        mean (np.ndarray): Array of N means in the shape N*M.
        cov (np.ndarray): Array of N covariance matrices in the shape NxMxM, where the covariance matrices are of dimension MxM.
        decomposition_type (str, optional): Type of decomposition to be implemented on the covariance matrix
        "svd" and "cholesky" are implemented. Take note that cholesky is faster but is only applicable for
        covariance matrices that are positive semi-definite / non-degenerative distributions.
        i.e. sigma !=0 for all dimensions of the noise
        . Defaults to "cholesky".
        check_valid (bool, optional): Determines if a check for validity is executed. Checks may cause significant
        slowdown in execution time Defaults to False.
        samples (int, optional): Number of points to sample from each covariance matrix. Defaults to 1.
        eps (float, optional): Due to rounding errors, a negative sigma might be
        calculated if sigma is suppose to be 0. Sigma with magnitude less that eps
        are rounded to z Defaults to 0.000000001. Only usd for svd decomposition as cholesky is not
        applicable for positive semi-definite

    Raises:
        NotImplementedError: Current LDL decomposition is not yet implemented

    Returns:
        np.ndarray: Array of sampled point of shape N*M*S whereby N is the number of covariance matrices.
        M is the dimension of the noise, S is the number of samples
    """

    valid_decompsitions = ["svd", "cholesky"]

    if decomposition_type not in valid_decompsitions:
        raise NotImplementedError("invalid decomposition type")
    no_cov = cov.shape[0]
    x = np.random.standard_normal((no_cov, mean.shape[-1], samples))
    if decomposition_type == "svd":
        (sigma, vec) = np.linalg.eig(cov)

        # Due to rounding errors, a negative sigma might be calculated if sigma is suppose to be 0
        sigma[np.abs(sigma) < eps] = 0
        if check_valid:
            assert np.all(vec > 0)

        rot_stretch_mat = vec * np.sqrt(sigma)[:, None, :]

    elif decomposition_type == "cholesky":
        rot_stretch_mat = np.linalg.cholesky(cov)

    elif decomposition_type == "ldl":
        raise NotImplementedError("LDLT decomposition has yet to be implemented")

    x_projected = np.matmul(rot_stretch_mat, x)
    x_projected += mean[:, :, None]

    return x_projected



import numpy as np
import open3d as o3d
import trimesh


class TSDFVolume(object):
    """Integration of multiple depth images using a TSDF."""

    def __init__(self, size, resolution, origin, with_colour=False):
        self.size = size
        self.resolution = resolution
        self.voxel_size = self.size / self.resolution
        self.origin = origin
        self.sdf_trunc = 4 * self.voxel_size
        self.with_colour = with_colour

        self.reset()

    def reset(self):
        if o3d.__version__ >= "0.13.0":
            o3d_integration = o3d.pipelines.integration
        else:
            o3d_integration = o3d.integration

        self._volume = o3d_integration.UniformTSDFVolume(
            length=self.size,
            resolution=self.resolution,
            sdf_trunc=self.sdf_trunc,
            color_type=(
                o3d_integration.TSDFVolumeColorType.RGB8
                if self.with_colour
                else o3d_integration.TSDFVolumeColorType.NoColor
            ),
            origin=self.origin[:3, 3],
        )

    def integrate(self, depth_img, intrinsic, extrinsic, colour_img=None):
        """
        Args:
            depth_img: The depth image.
            intrinsic: o3d.camera.PinholeCameraIntrinsic
            extrinsics: numpy array
        """

        colour_image = np.empty_like(depth_img) if colour_img is None else colour_img
        rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            o3d.geometry.Image(colour_image),
            o3d.geometry.Image(depth_img),
            depth_scale=1.0,
            depth_trunc=2.0,
            convert_rgb_to_intensity=False,
        )

        pose = np.linalg.inv(extrinsic)
        pose[:3, :3] = np.matmul(pose[:3, :3], self.origin[:3, :3])
        pose[:3, 3] -= self.origin[:3, 3]
        extrinsic = np.linalg.inv(pose)

        self._volume.integrate(rgbd, intrinsic, extrinsic)

    def get_grid(self):
        # TODO(mbreyer) very slow (~35 ms / 50 ms of the whole pipeline)
        shape = (1, self.resolution, self.resolution, self.resolution)
        tsdf_grid = np.zeros(shape, dtype=np.float32)
        voxels = self._volume.extract_voxel_grid().get_voxels()
        # vis_voxels = self._volume.extract_voxel_point_cloud()
        # o3d.visualization.draw_geometries([vis_voxels])
        for voxel in voxels:
            i, j, k = voxel.grid_index
            tsdf_grid[0, i, j, k] = voxel.color[0]
        return tsdf_grid

    def get_cloud(self):
        cloud = self._volume.extract_point_cloud()
        cloud = cloud.transform(self.origin)

        return cloud

    def get_mesh(self):
        mesh = self._volume.extract_triangle_mesh()

        mesh_smp = mesh.simplify_quadric_decimation(target_number_of_triangles=2000)
        mesh_smp.compute_vertex_normals(normalized=True)
        mesh_smp = mesh_smp.transform(self.origin)

        return mesh_smp

    def get_mesh_as_trimesh(self):
        mesh = self.get_mesh()

        vertices = np.asarray(mesh.vertices)
        triangles = np.asarray(mesh.triangles)
        normals = np.asarray(mesh.vertex_normals)

        # print(f"nvertices {vertices.shape} ntriangles {triangles.shape} nnormals {normals.shape}")
        tri_mesh = trimesh.Trimesh(vertices, triangles, vertex_normals=normals)

        return tri_mesh


class ScalableTSDFVolume(object):
    """Integration of multiple depth images using a TSDF."""

    def __init__(
        self, origin, depth_sampling_stride=4, resolution=9, with_colour=False
    ):
        self.origin = origin
        self.resolution = resolution
        scale = 2**self.resolution
        self.voxel_size = depth_sampling_stride / scale

        self.sdf_trunc = 0.04  # 4 * self.voxel_size

        self.with_colour = with_colour

    def reset(self):
        if o3d.__version__ >= "0.13.0":
            o3d_integration = o3d.pipelines.integration
        else:
            o3d_integration = o3d.integration

        self._volume = o3d_integration.ScalableTSDFVolume(
            voxel_length=self.voxel_size,
            volume_unit_resolution=self.resolution,
            sdf_trunc=self.sdf_trunc,
            color_type=(
                o3d_integration.TSDFVolumeColorType.RGB8
                if self.with_colour
                else o3d_integration.TSDFVolumeColorType.NoColor
            ),
        )

    def integrate(self, depth_img, intrinsic, extrinsic, colour_img=None):
        """
        Args:
            depth_img: The depth image.
            intrinsic: o3d.camera.PinholeCameraIntrinsic
            extrinsics: numpy array
        """

        colour_image = np.empty_like(depth_img) if colour_img is None else colour_img
        rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            o3d.geometry.Image(colour_image),
            o3d.geometry.Image(depth_img),
            depth_scale=1.0,
            depth_trunc=1.5,
            convert_rgb_to_intensity=False,
        )

        # pose = np.linalg.inv(extrinsic)
        # pose[:3, :3] = np.matmul(pose[:3, :3], self.origin[:3, :3])
        # pose[:3, 3] -= self.origin[:3, 3]
        # extrinsic = np.linalg.inv(pose)

        self._volume.integrate(rgbd, intrinsic, extrinsic)

    def get_cloud(self):
        return self._volume.extract_point_cloud()

    def get_mesh(self):
        return self._volume.extract_triangle_mesh()

    def get_mesh_as_trimesh(self):
        mesh = self.get_mesh()

        mesh = mesh.merge_close_vertices(0.001)
        mesh = mesh.remove_duplicated_vertices()
        mesh = mesh.remove_degenerate_triangles()

        vertices = np.asarray(mesh.vertices)
        triangles = np.asarray(mesh.triangles)
        normals = np.asarray(mesh.vertex_normals)

        # print(f"nvertices {vertices.shape} ntriangles {triangles.shape} nnormals {normals.shape}")
        tri_mesh = trimesh.Trimesh(vertices, triangles, vertex_normals=normals)

        return tri_mesh



from grip.agent import Agent
from grip.motion import PBBasicPlanner

import numpy as np


class DemoAgent(Agent):
    """Scripted agent that runs a pre-defined pick and place policy"""

    def __init__(self, **kwargs):
        self.env = kwargs.get("env")
        self.reward = kwargs.get("reward", None)

        self.planner = PBBasicPlanner(self.env.get_entity("robot"))

        self.action_idx = -1
        self.steps = 30

        self.finished = False
        self.gripper_actions = []
        self.arm_actions = []
        self.random_offset = np.zeros(3)

        self.env.world.register_keyhandler(ord("r"), self.reset)

    def reset(self):
        self.env.reset()

        self.generate_actions()

    def generate_actions(self, **obs):
        self.gripper_actions = []
        self.arm_actions = []
        self.finished = False
        self.action_idx = 0

        self.random_offset = obs.get("offset", self.random_offset)

        _, ee_ori = self.env.get_entity("robot").ee_pose
        obj_pos, _ = self.env.get_entity("object").pose
        grasp_pos = obj_pos + np.array([0.0, 0.0, 0.05])
        grasp_ori = ee_ori

        place_pos = [0.49892282, 0.3771686, 0.16378036]
        place_ori = [-0.653, 0.271, 0.271, 0.653]

        self.compute_trajectory(grasp_pos, grasp_ori, place_pos, place_ori)

    def compute_trajectory(self, grasp_pos, grasp_ori, place_pos, place_ori):
        grasp_arm_actions, grasp_gripper_actions = self.get_grasp_actions(
            grasp_pos, grasp_ori
        )

        place_arm_actions, place_gripper_actions = self.get_placement_actions(
            place_pos, place_ori, initial_conf=grasp_arm_actions[-1]
        )

        self.arm_actions = grasp_arm_actions + place_arm_actions
        self.gripper_actions = grasp_gripper_actions + place_gripper_actions

    def observe(self, obs):
        self.finished = self.action_idx >= len(self.arm_actions)

        if self.action_idx < 0:
            self.generate_actions(**obs)

        # log.info("OBS: {}".format(obs))

        return self.finished

    def act(self):
        action = {}
        if not self.finished:
            action = {
                "joint_position_cmd": self.arm_actions[self.action_idx],
                "gripper_cmd": self.gripper_actions[self.action_idx],
            }

            self.action_idx += 1

        return action

    def get_grasp_actions(self, grasp_pos, grasp_ori):
        obstacles = [self.env.get_entity(name).id for name in ["tray", "plane"]]
        disabled_collisions = [self.env.get_entity("object").id]

        grasp_path = self.planner.plan_cartesian_goal(
            (grasp_pos, grasp_ori),
            obstacles=obstacles,
            disabled_collisions=disabled_collisions,
        )

        gripper_actions = np.zeros(len(grasp_path))
        gripper_actions = np.hstack([gripper_actions, -np.ones(self.steps) * 0.05])

        grasp_path = grasp_path + [grasp_path[-1]] * self.steps

        lift_pos = grasp_pos
        lift_pos[2] += 0.20
        lift_path = self.planner.plan_cartesian_goal(
            (lift_pos, grasp_ori), obstacles=[], initial_conf=grasp_path[-1]
        )

        gripper_actions = np.hstack([gripper_actions, -np.ones(len(lift_path)) * 0.05])

        grasp_path = grasp_path + lift_path

        return grasp_path, gripper_actions.tolist()

    def get_placement_actions(self, place_pos, place_ori, initial_conf=None):
        obstacles = [self.env.get_entity(name).id for name in ["tray", "plane"]]

        pre_place_path = self.planner.plan_joint_goal(
            self.env.panda.home_positions,
            obstacles=obstacles,
            initial_conf=initial_conf,
        )

        placement_path = self.planner.plan_cartesian_goal(
            (place_pos, place_ori), obstacles=[], initial_conf=pre_place_path[-1]
        )

        placement_path = pre_place_path + placement_path

        gripper_actions = np.zeros(len(placement_path))
        gripper_actions = np.hstack([gripper_actions, np.ones(self.steps) * 0.05])

        placement_path += [placement_path[-1]] * self.steps

        ee_pos, ee_ori = self.env.panda.forward_kinematics(placement_path[-1])
        ee_pos[0] = ee_pos[0] - 0.15
        release_path = self.planner.plan_cartesian_goal(
            (ee_pos, ee_ori), obstacles=[], initial_conf=placement_path[-1]
        )

        gripper_actions = np.hstack(
            [gripper_actions, np.zeros(len(release_path)) * 0.05]
        )

        placement_path += release_path

        gripper_actions = np.hstack([self.gripper_actions, gripper_actions])

        return placement_path, gripper_actions.tolist()



from dataclasses import dataclass
import pybullet as p
from ..io import log
from .entity import Entity
import atexit
import sys, os, gc
from typing import List, Callable, Optional, Union
import functools, inspect
import pybullet_data
import pkgutil


class BulletClient:
    def __init__(self, phys_id, pid):
        self._pid = pid
        self._client = phys_id

    def __getattr__(self, name):
        attribute = getattr(p, name)
        if inspect.isbuiltin(attribute):
            attribute = functools.partial(attribute, physicsClientId=self._client)
        if name == "disconnect":
            self._client = -1
        return attribute


class BulletWorld:
    """
    The BulletWorld represents an interface and connection to a Bullet physics simulation.

    """

    def __init__(self, **kwargs):
        """
        An object of this class is reponsible for esblishing communication to a new or existing bullet physics engine.

        Args:
            phys_opt (str): type of connection to bullet, it can assume one of the following values ['direct', 'gui', 'shared_memory', 'udp', 'tcp'].
                "direct": physics engine world is created headless, without GUI.

                "gui": physics engine world is created with opengl rendering as GUI.

                "shared_memory": connects to an existing physics server via shared memory.

                "udp": connects to an already existing physics server with the UDP protocol
                    you must provide hostname and port if you choose this option.

                "tcp": physics engine world is created connecting to an already existing physics server with the TCP communication.
                    You must provide hostname and port if you choose this option.

            gravity (list[float], optional): gravity vector. Defaults to [0, 0, -9.81].
            real_time_sim (bool, optional): If you enable the real-time simulation, you don't need to call 'world.step()'.
                Has no effect if world phys_opt is connected as 'direct', in this case you will have to call world.step(). Defaults to True.
            deformable (bool, optional): configure world to support deformable physics. Defaults to False.
            solver_iterations (int, optional): internal parameter representing maximum number of constraint solver iterations.Defaults to 240.
            timestep (float, optional): time step used for physics evolution. Defaults to 1/240.
            reset_at_init (bool, optional): if this is true, the method reset() will be called as this object is created.

        """

        self.phys_opt = kwargs.get("phys_opt", "direct")
        self.hostname = kwargs.get("hostname", "localhost")
        self.port = kwargs.get("port", 1234)
        self.gravity = kwargs.get("gravity", [0, 0, -9.81])
        self.real_time_sim = kwargs.get("real_time_sim", True)
        self.deformable = kwargs.get("deformable", False)
        self.solver_iterations = kwargs.get("solver_iterations", 240)
        self.timestep = kwargs.get("timestep", 1.0 / 240.0)
        self.reset_at_init = kwargs.get("reset_at_init", True)
        self.texture_randomiser = None
        self.phys_id = (
            -1
        )  # invalid phys_id initialised by default (-1 means disconected)
        options = "--background_color_red=0.2 --background_color_blue=0.2 --background_color_green=0.2"

        flag = p.DIRECT
        if self.phys_opt == "shared_memory":
            flag = p.SHARED_MEMORY
        elif self.phys_opt in ["gui", "GUI"]:
            flag = p.GUI
        elif self.phys_opt in ["gui_server", "GUI_SERVER"]:
            flag = p.GUI_SERVER
        elif self.phys_opt in ["UDP", "udp"]:
            flag = p.UDP
        elif self.phys_opt in ["TCP", "tcp"]:
            flag = p.TCP
        elif self.phys_opt == "egl":
            flag = p.SHARED_MEMORY_SERVER

        if self.phys_opt in [
            "gui",
            "GUI",
            "gui_server",
            "direct",
            "DIRECT",
            "shared_memory",
            "egl",
        ]:
            self.phys_id = p.connect(
                flag, options=options
            )  # (pb.DIRECT), pb.SHARED_MEMORY, pb.GUI #"--opengl2"
            if self.reset_at_init:
                self.reset()

        elif self.phys_opt in ["UDP", "udp"]:
            self.phys_id = p.connect(flag, self.hostname, self.port)
            p.setTimeOut(10)
        elif self.phys_opt in ["TCP", "tcp"]:
            log.info("Trying to connect to physics")
            self.phys_id = p.connect(flag, self.hostname, self.port)
            p.setTimeOut(10)

        if self.phys_id < 0:
            log.info("Failed to connect to physics engine")

        self.state_id = None

        self._pid = os.getpid()

        self.client = BulletClient(self.phys_id, self._pid)

        atexit.register(self.on_shutdown)

        self.entity_list = []
        self.keyboard_handlers = {ord("q"): self.shutdown}

        if self.phys_opt != "shared_memory":
            p.setAdditionalSearchPath(pybullet_data.getDataPath())

        self.image_textures = {}

        self._shutdown_callbacks = []

        if self.phys_opt == "egl":
            loaded_egl = self.load_egl_plugin()

            if loaded_egl:
                log.info("EGL plugin loaded successfully!")
            else:
                log.warning("EGL plugin failed to load")

    def load_egl_plugin(self) -> bool:
        """Loads EGL plugin

        Returns:
            (bool): whether or not the plugin was loaded successfully
        """

        egl = pkgutil.get_loader("eglRenderer")
        if egl:
            self.plugin_id = p.loadPlugin(
                egl.get_filename(), "_eglRendererPlugin", physicsClientId=self.phys_id
            )
        else:
            self.plugin_id = p.loadPlugin(
                "eglRendererPlugin", physicsClientId=self.phys_id
            )
        log.info(f"pluginId={self.plugin_id}")

        return self.plugin_id > 0

    def unload_egl_plugin(self) -> None:
        """Unloads EGL plugin"""

        if self.plugin_id > 0:
            p.unloadPlugin(self.plugin_id, physicsClientId=self.phys_id)

    def __del__(self):
        """Clean up connection if not already done."""

        # log.info("Clearing world resources! Current phys_id: {} pid: {}, os.getpid(): {}".format(self.id, self._pid, os.getpid()))
        self.disconnect()

    def disconnect(self) -> None:
        gc.collect()
        if self.id >= 0 and self._pid == os.getpid():
            try:
                # log.info("Attempting disconnecting this engine")

                p.removeAllUserParameters(physicsClientId=self.id)
                p.removeAllUserDebugItems(physicsClientId=self.id)

                self.clear_entities()

                p.disconnect(physicsClientId=self.id)
                self.phys_id = -1
                # log.info("Disconnecting this engine was successful")
            except p.error:
                pass
                # log.info("Failed to clean: {}".format(e))

    @property
    def id(self):
        """
        int: The unique identifier of this physics simulation connection.
        """

        return self.phys_id

    def step(self) -> None:
        """
        Evolves this simulation by one step step.
        """

        if self.state_id is None:
            self.save_state()

        p.stepSimulation(physicsClientId=self.id)

        self.handle_keyboard()

    def reset(self, load_plane: bool = False) -> None:
        """
        Completely resets this world. Starts a fresh world without any entities.
        """

        if self.deformable:
            p.resetSimulation(p.RESET_USE_DEFORMABLE_WORLD, physicsClientId=self.id)
        else:
            p.resetSimulation(physicsClientId=self.id)

        if self.real_time_sim:
            p.setRealTimeSimulation(1, physicsClientId=self.id)
        else:
            p.setTimeStep(self.timestep, physicsClientId=self.id)

        p.setPhysicsEngineParameter(
            numSolverIterations=self.solver_iterations, physicsClientId=self.id
        )

        p.setGravity(*self.gravity, physicsClientId=self.id)

        if load_plane:
            p.loadURDF(
                os.path.join(pybullet_data.getDataPath(), "plane.urdf"), [0, 0, 0]
            )

    def set_real_time(self, real_time_sim: bool) -> None:
        """
        Sets real-time sim flag

        real_time_sim: whether or not to step the simulation according to its real-time-clock (RTC) on a separate thread.
        """

        self.real_time_sim = real_time_sim

        p.setRealTimeSimulation(self.real_time_sim, physicsClientId=self.id)

    def set_gravity(self, gravity: List[float]) -> None:
        """
        Sets gravity of the world

        Args:
            gravity (list[float]): three dimensional gravity vector.

        """

        self.gravity = gravity

        p.setGravity(*self.gravity, physicsClientId=self.id)

    def remove_debug_item(self, item_id: int) -> None:
        """
        Removes a debug item from the simulation identified by its item_id.

        Args:
            item_id (int): item unique ID to be removed.

        """

        p.removeUserDebugItem(item_id, physicsClientId=self.id)

    def shutdown(self) -> None:
        """
        Exists and closes this process. Internally, it calls sys.exit().

        """

        sys.exit()

    def on_shutdown(self) -> None:
        """
        Removes objects, GUI elements, entities and disconnects from the physics server

        """

        if self.is_connected():
            self.disconnect()

            # log.info("Exiting GRIP. Bye!")

        for cb in self._shutdown_callbacks:
            cb()

    def shutdown_register(self, shutdown_callback: Callable[[], None]):
        assert callable(shutdown_callback)

        self._shutdown_callbacks.append(shutdown_callback)

    def setDebugCamera(
        self, view_dist=1, view_yaw=0, view_pitch=1.0, view_target_pos=None, **kwargs
    ) -> None:
        """
        Sets GUI view camera, the observer view in the GUI.

        Args:
            view_dist (float): distance from eye to camera target position
            view_yaw (float): camera yaw angle (in degrees) left/right
            view_pitch (float): camera pitch angle (in degrees) up/down
            view_target_pos (list[float]): cameraTargetPosition is the camera focus point in 3D space.

        """
        if view_target_pos is None:
            view_target_pos = [0, 0, 0.5]
        p.resetDebugVisualizerCamera(
            view_dist, view_yaw, view_pitch, view_target_pos, physicsClientId=self.id
        )

    def hide_gui(self) -> None:
        """
        Hides GUI panels from debug visualiser
        """
        p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0, physicsClientId=self.id)

    def save_state(self) -> None:
        """
        Takes a snapshot and saves the current state of the simulation in-memory.
        This way the simulation can be efficiently restarted by calling restore_state()
        """

        if self.state_id is not None:
            p.removeState(self.state_id, physicsClientId=self.id)

        self.state_id = p.saveState(physicsClientId=self.id)

    def restore_state(self) -> None:
        """
        Restores the simulation to a previous state aved by the most recent call to save_state()
        """

        if self.state_id is not None:
            p.restoreState(self.state_id, physicsClientId=self.id)

    def is_connected(self) -> bool:
        """
        Tells if this world is still connected to the physics server

        Returns:
            bool: connection_status
        """

        return p.isConnected(physicsClientId=self.id)

    def set_data(self, entity: Entity, key: str, value: str) -> None:
        """
        Attaches a key-value pair to the provided entity. Can be used for storing a variable globally
        across shared memory instances of pybullet.

        Args:
            entity (Entity): the object to which the data should be attached
            key (str): the key to be used for accessing the object
            value (str): the value to be stored for the key
        """
        p.addUserData(entity.id, key, value, physicsClientId=self.id)

    def add_data(self, key: str, value: str):
        """
        Attaches a key-value pair to a new dummy object. Can be used for storing a variable globally
        across shared memory instances of pybullet. Creates a dummy object and attaches value using
        :py:meth:`set_data`.

        Args:
            key (str): the key to be used for accessing the object
            value (str): the value to be stored for the key
        """
        shape_id = p.createVisualShape(
            p.GEOM_BOX,
            halfExtents=[0.01] * 3,
            rgbaColor=[0, 0, 0, 0],
            physicsClientId=self.id,
        )
        box_id = p.createMultiBody(
            0, -1, shape_id, [0] * 3, [0, 0, 0, 1], physicsClientId=self.id
        )

        @dataclass
        class dummy:
            id: int = box_id

        self.register_entity(dummy)
        self.set_data(dummy, "package_name", "panda_flex_dev")

    def get_data(self, query_key: str) -> Optional[str]:
        """
        Retrieves data from shared memory using provided key. Assumes key is unique across all enitities.
        (See :py:meth:`add_data` and :py:meth:`set_data`)

        Args:
            key (str): the key used for storing the object

        Returns:
            value (str): the value stored for the key. Returns None on failure.
        """
        bodies_uids = self.get_bodies_uids()
        for buid in bodies_uids:
            for i in range(p.getNumUserData(buid, physicsClientId=self.id)):
                (
                    userDataId,
                    key,
                    bodyId,
                    linkIndex,
                    visualShapeIndex,
                ) = p.getUserDataInfo(buid, i, physicsClientId=self.id)
                key = key.decode("utf-8")
                if key == query_key:
                    return p.getUserData(userDataId, physicsClientId=self.id).decode(
                        "utf-8"
                    )

        return None

    def get_bodies_uids(self) -> List[int]:
        """
        Get uids of all objects in the sim world.

        Returns:
            bodies_uids (List[int]): uids of all objects in the simulation.
        """
        n_bodies = p.getNumBodies(physicsClientId=self.id)

        bodies_uids = []
        for i in range(n_bodies):
            buid = p.getBodyUniqueId(i, physicsClientId=self.id)
            bodies_uids.append(buid)

        return bodies_uids

    def get_body_by_name(self, name: str) -> int:
        """
        Retrieves a pre-existing body unique identifier from the simulation searched by its name
        The name is typically given in the objects' URDF definition.

        Args:
            name (str): Name of the body as defined in its URDF.

        Returns:
            body_idx (int): unique identifier of the body in this world
        """

        n_bodies = p.getNumBodies(physicsClientId=self.id)

        body_idx = -1
        for i in range(n_bodies):
            buid = p.getBodyUniqueId(i, physicsClientId=self.id)
            info = p.getBodyInfo(buid, physicsClientId=self.id)
            log.info("BodyInfo {}: {}".format(i, info))

            body_name = info[1].decode("utf-8")

            if body_name == name:
                body_idx = buid
                break

        if body_idx < 0:
            log.debug("Body name {} not found")

        return body_idx

    def get_body_names(self) -> List[str]:
        """
        Retrieves all pre-existing bodies names that are given in the objects' URDF definition.

        Returns:
            body_names (list[str]): list of names for each  in this world
        """

        n_bodies = p.getNumBodies(physicsClientId=self.id)

        bodies = []
        for i in range(n_bodies):
            buid = p.getBodyUniqueId(i, physicsClientId=self.id)
            info = p.getBodyInfo(buid, physicsClientId=self.id)
            log.info("BodyInfo {}: {}".format(buid, info))

            body_name = info[1].decode("utf-8")

            bodies.append(body_name)

        return bodies

    def register_keyhandler(self, key_code: int, func: Callable[[], None]) -> None:
        """
        Registers a callback function for a given key_code

        Args:
            key_code (int): int
                key code, can be obtained using builtin function ord(..)
            func (callable):
                a callback function that does not receive any parameters.
                the signature of the callback function should be `def fun() -> None`

        Returns:
            body_idx (int): unique identifier of the body in this world

        Examples:
            >>> world.register_keyhandler(ord('a'), callback_fun)
        """

        self.keyboard_handlers[key_code] = func

    def register_entity(self, entity: Union[Entity, List[Entity]]) -> None:
        """
        Registers an entity or list of entities to this world for memory management purposes
        This allows entities registered to this world to be removed on shutdown

        This is useful if this world is connected to a central physics server,
        which in turn the physics server may be connected to many other worlds.
        When each world adds its entities, each world then becomes responsible for removing them
        after they are shutdown.


        Args:

            entity (Union[Entity,List[Entity]]): entity or entities to be registed, individual entities must inherit from grip.robot.Entity
                and implement the basic grip.robot.Entity interface.

        """

        if isinstance(entity, list):
            self.entity_list.extend(entity)
        else:
            self.entity_list.append(entity)

        p.configureDebugVisualizer(
            p.COV_ENABLE_MOUSE_PICKING, 1, physicsClientId=self.id
        )

    def handle_keyboard(self) -> None:
        """
        Handles keyboard events that were register with register_keyhandler()

        """

        keys = p.getKeyboardEvents(physicsClientId=self.id)

        reg_keys = self.keyboard_handlers.keys()

        for key in keys:
            if keys[key] & p.KEY_WAS_TRIGGERED:
                log.info("Key {} was pressed.".format(key))

        for key in reg_keys:
            if key in keys and keys[key] & p.KEY_WAS_TRIGGERED:
                self.keyboard_handlers[key]()

    def clear_entities(self) -> None:
        """
        Clear all entities registered to this world

        """

        for entity in self.entity_list:
            p.removeBody(entity.id, physicsClientId=self.id)
        self.entity_list = []

    def remove_entity(self, entity_id: int) -> bool:
        """
        Clear a specific entity provided its unique identifier

        Args:
        entity_id (int): unique identifier of the entity

        Returns:
            bool: whether or not the removal was successfull.
        """

        success = False
        try:
            p.removeBody(entity_id, physicsClientId=self.id)
            success = True
            if entity_id in self.entity_list:
                self.entity_list.remove(entity_id)

        except Exception as e:
            log.warning(
                f"Problem when trying to remove entity id: {entity_id}. Message {e}"
            )
        finally:
            return success

    def texture(self, texture_path: str) -> int:
        """
        Retrieves a texture's unique identifier (uid) given its path.
        If a texture was already loaded in memory with the same path, then the existing texture id will be retrived.
        Otherwise, the texture will be loaded and its uid will be returned.


        Args:
            texture_path (str): full path of texture file (typically a .png file)
        Returns:
            int: unique identifier of loaded texture
        """

        if self.image_textures.get(texture_path, None) is None:
            self.add_texture(texture_path)

            log.debug("Loading to memory!!")
        else:
            log.debug(f"Hit!! {texture_path}")

        return self.image_textures[texture_path]

    def add_texture(self, texture_path: str) -> None:
        """
        Loads texture located in texture_path to memory and adds to this world's texture dictionary.


        Args:
            texture_path (str): full path of texture file (typically a .png file)
        """

        self.image_textures[texture_path] = p.loadTexture(
            texture_path, physicsClientId=self.id
        )

        log.info(f"Loaded texture {texture_path}: {self.image_textures[texture_path]}")



from ..io import get_data_path, log
from ..robot import BulletRobot, BulletObject, BulletWorld
import pybullet as p
import numpy as np
from typing import Union, Tuple, List
from dataclasses import dataclass

base_dir = get_data_path()


class TextureRandomiser:
    """An object of this class is able to randomise the texture of BulletObject and BulletRobot instances"""

    def __init__(self, world: BulletWorld, texture_paths: List[str]):
        """
        Constructs a TextureRandomiser object.
        It can be used for to randomising the texture of BulletObject and BulletRobot instances.

        Args:
            world: a parent bullet world instance where entities to be randomised exist
            texture_paths: a list of texture paths to be chosen from

        """

        self.world = world
        self.textures_paths = texture_paths
        self.texture_count = len(texture_paths)

        self.world.texture_randomiser = self

        log.info(f"Loaded textures: {self.world.image_textures}")

    def randomise(self, entity: Union[BulletObject, BulletRobot]) -> None:
        """Sets texture of given entity to be a uniformly random one"""

        if isinstance(entity, BulletObject):
            tid = self.random_texture_id()
            p.changeVisualShape(
                entity.id, -1, textureUniqueId=tid, physicsClientId=self.world.id
            )
        elif isinstance(entity, BulletRobot):
            tid = self.random_texture_id()

            shape_data = p.getVisualShapeData(entity.id, physicsClientId=self.world.id)

            for shape in shape_data:
                p.changeVisualShape(
                    entity.id,
                    shape[1],
                    textureUniqueId=tid,
                    physicsClientId=self.world.id,
                )

    def random_texture_id(self) -> int:
        """
        gets random texture unique identifier (chosen uniformly)

        Returns:

            int: uniform randomly selected unique identifier of loaded texture (assuming texture already existing in memory)

        """

        tex_idx = np.random.randint(0, self.texture_count)

        tid = self.world.texture(self.textures_paths[tex_idx])

        return tid


@dataclass
class LightingParameters:
    """This class represents the basic phong illumination model parameters"""

    light_direction: np.ndarray = np.ones(3)
    light_colour: np.ndarray = np.ones(3)
    specular_coef: int = 1
    diffuse_coef: int = 1
    ambient_coef: int = 1


class LightingRandomiser:
    """Randomises phong illumination parameters"""

    DEFAULT_PARAMS = LightingParameters()

    def __init__(
        self,
        pos_range: List[Tuple[float, float]] = [[-1, 1], [-1, 1], [1, 2.5]],
        colour_range: List[Tuple[float, float]] = [[0, 1], [0, 1], [0, 1]],
        specular_range: Tuple[float, float] = [0, 1],
        diffuse_range: Tuple[float, float] = [0, 1],
        ambient_range: Tuple[float, float] = [0, 1],
    ):
        self._pos_range = pos_range
        self._colour_range = colour_range
        self._specular_range = specular_range
        self._diffuse_range = diffuse_range
        self._ambient_range = ambient_range

    def randomise(self) -> LightingParameters:
        """Returns randomised lighting parameters

        Returns:
            (LightingParameters): randomised phong illumination model parameters
        """

        light_direction = [
            np.random.uniform(self._pos_range[0][0], self._pos_range[0][1]),
            np.random.uniform(self._pos_range[1][0], self._pos_range[1][1]),
            np.random.uniform(self._pos_range[2][0], self._pos_range[2][1]),
        ]
        light_color = [
            max(
                0,
                min(
                    1,
                    np.random.uniform(
                        self._colour_range[0][0], self._colour_range[0][1]
                    ),
                ),
            ),
            max(
                0,
                min(
                    1,
                    np.random.uniform(
                        self._colour_range[1][0], self._colour_range[1][1]
                    ),
                ),
            ),
            max(
                0,
                min(
                    1,
                    np.random.uniform(
                        self._colour_range[2][0], self._colour_range[2][1]
                    ),
                ),
            ),
        ]
        specular_coeff = max(
            0,
            min(1, np.random.uniform(self._specular_range[0], self._specular_range[1])),
        )
        diffuse_coeff = max(
            0, min(1, np.random.uniform(self._diffuse_range[0], self._diffuse_range[1]))
        )
        ambient_coeff = max(
            0, min(1, np.random.uniform(self._ambient_range[0], self._ambient_range[1]))
        )
        return LightingParameters(
            light_direction, light_color, specular_coeff, diffuse_coeff, ambient_coeff
        )



import abc

from typing import Protocol
from grip.robot import BulletWorld
from typing import Any


class Env(abc.ABC):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def step(self, action):
        """Run one timestep of the environment's dynamics. When end of
        episode is reached, you are responsible for calling `reset()`
        to reset this environment's state.

        Accepts an action and returns a tuple (observation, reward, done, info).

        Args:
            action (object): an action provided by the agent

        Returns:
            observation (object): agent's observation of the current environment
            reward (float) : amount of reward returned after previous action
            done (bool): whether the episode has ended, in which case further step() calls will return undefined results
            info (dict): contains auxiliary diagnostic information (helpful for debugging, and sometimes learning)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self):
        """Resets the state of the environment and returns an initial observation.

        Returns:
            observation (object): the initial observation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def render(self, mode="human"):
        """Renders the environment.

        The set of supported modes varies per environment. (And some
        environments do not support rendering at all.) By convention,
        if mode is:

        - human: render to the current display or terminal and
          return nothing. Usually for human consumption.
        - rgb_array: Return an numpy.ndarray with shape (x, y, 3),
          representing RGB values for an x-by-y pixel image, suitable
          for turning into a video.
        - ansi: Return a string (str) or StringIO.StringIO containing a
          terminal-style text representation. The text can include newlines
          and ANSI escape sequences (e.g. for colors).

        Note:
            Make sure that your class's metadata 'render.modes' key includes
              the list of supported modes. It's recommended to call super()
              in implementations to use the functionality of this method.

        Args:
            mode (str): the mode to render with

        Example:

        class MyEnv(Env):
            metadata = {'render.modes': ['human', 'rgb_array']}

            def render(self, mode='human'):
                if mode == 'rgb_array':
                    return np.array(...) # return RGB frame suitable for video
                elif mode == 'human':
                    ... # pop up a window and render
                else:
                    super(MyEnv, self).render(mode=mode) # just raise an exception
        """
        raise NotImplementedError

    def close(self):
        """Override close in your subclass to perform any necessary cleanup.

        Environments will automatically close() themselves when
        garbage collected or when the program exits.
        """

    @abc.abstractmethod
    def add(self, entity):
        """Override add in your subclass to add entities to this environment."""

    @abc.abstractmethod
    def remove(self, entity):
        """Override add in your subclass to remove enities to this environment"""


class RobotEnv(Protocol):
    """A generic robot environment interface.

    It has a init method which receives a world and any extra needed parameters.
    The init method should implement a world initialisation, e.g. instantiating objects and positioning them at desired locations

    """

    def init(self, parent_world: BulletWorld, *args: Any, **kwargs: Any) -> None:
        """Initialises this environment"""



from typing import Callable, Type

from ..io import import_module
from .gripper_interface import GripperInterface


class EndEffectorType(GripperInterface):
    # Type name for real hardware interface
    HW_INTERFACE = "path.to.package:ClassName"

    @classmethod
    def get_hw_interface(cls):
        module_path = cls.HW_INTERFACE.split(":")[0]
        class_name = cls.HW_INTERFACE.split(":")[1]

        module = import_module(module_path)

        assert (
            module is not None
        ), f"Importing module {module_path} has failed. Are you sure it exists?"

        return getattr(module, class_name)


class EndEffectorRegistry:
    registry = {}

    @classmethod
    def register(cls, name: str) -> Callable:
        """
        Class method to register gripper classes to the internal gripper registry.

        Args:
            name: The functor name
        """

        def inner_wrapper(gripper_class: EndEffectorType) -> Callable:
            if name in cls.registry:
                raise RuntimeError(f"Error: {name} already exists in {cls.__name__}.")
            cls.registry[name] = gripper_class

            return gripper_class

        return inner_wrapper

    @classmethod
    def make(cls, name: str, *args, **kwargs) -> EndEffectorType:
        """
        Class method used to construct a gripper registered with a given name

        Args:
            name: the name of the gripper type
        Returns:
            (EndEffectorType): returns an instance of the end-effector registered with the given name.
        """
        gripper_class = cls.get(name)

        return gripper_class(*args, **kwargs)

    @classmethod
    def get(cls, name: str) -> Type[EndEffectorType]:
        """
        Class method used to get the end-effector type registered with given name.

        Args:
            name: the name of the gripper type
        Returns:
            (Type[EndEffectorType]): returns the class type of the gripper registered with given name.
        """

        if name not in cls.registry:
            raise RuntimeError(f"Error: {name} not registered in {cls.__name__}.")

        gripper_class = cls.registry[name]

        return gripper_class

    @classmethod
    def is_registered(cls, name: str) -> bool:
        """
        Checks if a gripper type with given name is already registered in this registry.

        Args:
            name: the name of the gripper type
        Returns:
            (bool): returns whether or not the type name has been already registered.
        """

        return name in cls.registry


EndEffectorRegistry.registry["none"] = None



#!/usr/bin/env python3

import os
import numpy as np

try:
    import rclpy
    from rclpy.node import Node

except ImportError:
    raise RuntimeError("Unable to run example without ros2 installed.")

import grip
import grip.ros
import pybullet_data as pd

import cv2


class ROSCameraExample(Node):
    def __init__(self, fps=10):
        super().__init__("rgbd_camera_example_node")

        # Creates bullet world with GUI
        self.world = grip.robot.BulletWorld(phys_opt="gui")

        # Creates ground plane object
        self.plane = grip.robot.BulletObject(
            urdf_file=os.path.join(pd.getDataPath(), "plane.urdf"), position=[0, 0, 0]
        )

        # Creates ROSCamera object
        self.camera = grip.ros.ROSCamera(
            cid=self.world.id,
            position=np.array([0, 0, 1.5]),
            orientation=grip.math.rpy2quaternion([0.0, np.pi, 0.0]),
            parent_node=self,
            image_topic="/hand_camera/color/image_raw",
            image_info_topic="/hand_camera/color/camera_info/",
            depth_topic="/hand_camera/aligned_depth_to_color/image_raw",
            depth_info_topic="hand_camera/aligned_depth_to_color/camera_info",
            base_frame_id="hand_camera_link",
            sim=False,
        )

        self.timer = self.create_timer(1 / fps, self.on_update)

    def on_update(self):
        if not self.camera.is_ready:
            grip.io.log.info("Camera is not ready.")
            return

        rgb, depth = self.camera.obs()

        bgr = cv2.cvtColor(rgb, cv2.COLOR_RGB2BGR)

        depth_colormap = cv2.applyColorMap(
            # convert to mm
            cv2.convertScaleAbs(depth * 1e3, alpha=0.03),
            cv2.COLORMAP_JET,
        )

        images = np.hstack([bgr, depth_colormap])
        cv2.imshow("RGB-Depth", images)

        cv2.waitKey(1)


def main(args=None):
    rclpy.init(args=args)

    ros_camera_example = ROSCameraExample()

    rclpy.spin(ros_camera_example)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    ros_camera_example.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()



import abc
from typing import Protocol


class RobotInterface(abc.ABC):
    """
    Command methods with side effects
    """

    @abc.abstractmethod
    def exec_cartesian_cmd(self, cmd_p, cmd_q):
        """
        Not implemented
        """
        raise NotImplementedError("exec_cartesian_cmd: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def exec_cartesian_delta_cmd(self, cmd_dp, cmd_dq):
        """
        Not implemented
        """
        raise NotImplementedError(
            "exec_cartesian_delta_cmd: NO EFFECT, NOT IMPLEMENTED"
        )

    @abc.abstractmethod
    def exec_position_cmd(self, cmd):
        """
        Not implemented
        """

        raise NotImplementedError("exec_position_cmd: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def exec_position_cmd_delta(self, cmd):
        """
        Not implemented
        """
        raise NotImplementedError("exec_position_cmd_delta: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def exec_velocity_cmd(self, cmd):
        """
        Not implemented
        """
        raise NotImplementedError("exec_velocity_cmd: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def exec_torque_cmd(self, cmd):
        """
        Not implemented
        """
        raise NotImplementedError("exec_torque_cmd: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def move_to_home(self):
        """
        Not implemented
        """
        raise NotImplementedError("move_to_home: NO EFFECT, NOT IMPLEMENTED")

    """
    Get methods
    """

    @property
    @abc.abstractmethod
    def q_mean(self):
        """
        Not implemented, should return the mean between joint limits
        """
        raise NotImplementedError("q_mean: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def n_joints(self):
        """
        Not implemented, should return the number of joints of this robot, may differ to n_cmd()
        """
        raise NotImplementedError("n_joints: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def angles(self):
        """
        Returns current joint angles measured by encoders
        """
        raise NotImplementedError("angles: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def joint_limits(self):
        """
        Returns joint limits
        """
        raise NotImplementedError("joint_limits: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def joint_names(self):
        """
        Returns list of joint names for this robot
        """
        raise NotImplementedError("joint_names: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def joint_velocities(self):
        """
        Returns joint velocities measurements values for this robot
        """
        raise NotImplementedError("joint_velocities: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def joint_efforts(self):
        """
        Returns joint effort measurements values for this robot
        """
        raise NotImplementedError("joint_efforts: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def ee_pose(self):
        """
        Not implemented, should return end-effector pose
        """
        raise NotImplementedError("ee_pose: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def ee_velocity(self):
        """
        Not implemented, should return end-effector velocity
        """
        raise NotImplementedError("ee_velocity: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def base_pose(self):
        """
        Not implemented, should return robot base pose
        """
        raise NotImplementedError("base_pose: NO EFFECT, NOT IMPLEMENTED")

    @property
    @abc.abstractmethod
    def base_velocity(self):
        """
        Not implemented, should return robot spatial velocity
        """
        raise NotImplementedError("base_velocity: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def forward_kinematics(self, joint_angles=None):
        """
        Not implemented, should return the forward kinematics over joint_angles
        """
        raise NotImplementedError("forward_kinematics: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def inverse_kinematics(self, position, orientation=None):
        """
        Not implemented, should attempt to solve inverse kinematics
        for the given position and orientation
        """
        raise NotImplementedError("inverse_kinematics: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def jacobian(self, joint_angles=None):
        """
        Not implemented, should return the jacobian J(q) for the given joint angles
        """
        raise NotImplementedError("jacobian: NO EFFECT, NOT IMPLEMENTED")

    @abc.abstractmethod
    def inertia(self, joint_angles=None):
        """
        Not implemented, should return the inertia tensor for the given joint angles M(q)
        """
        raise NotImplementedError("inertia: NO EFFECT, NOT IMPLEMENTED")



#!/usr/bin/env python3

import numpy as np
import open3d as o3d
from scipy.spatial import KDTree
from scipy.spatial.transform import Rotation as R

from ..robot.gui import visualizePcd


class RGBD2PCDParser(object):
    """
    This is a class that extract point cloud and infer other required information from
    rgbd information and

    :param int width: image width
    :param int height: image height
    :param float fx: focal length in x
    :param float fy: focal length in y
    :param float cx: x coordinate for camera principle point
    :param float cy: y coordinate for camera principle point
    :param float camera_near: near plane distance
    :param float camera_far: far plane distance
    :param float depth_scale: scale for current depth value wrt to 1mm
    """

    def __init__(
        self, width, height, fx, fy, cx, cy, camera_near, camera_far, depth_scale
    ):
        self.camera_position = [0, 0, 0]
        self.near = camera_near
        self.far = camera_far
        self.depth_scale = depth_scale
        self.depth_truncate = self.near * 100 + 1
        self.camera_intrinsic = o3d.camera.PinholeCameraIntrinsic(
            width=width, height=height, fx=fx, fy=fy, cx=cx, cy=cy
        )

        self.camera_extrinsic = np.eye(4)

    def extract_mask(self, seg_image, obj_id):
        """
        Extract a mask on target object for RGBD images

        :param nd.ndarray seg_image: input segmentation image from camera image
                                     rendered at the same time with rgb image
        :param [int] obj_id: list of object for target object id in simulation

        :return: A mask as a (width, height) array of boolean value
        :rtype: np.ndarray
        """
        if obj_id == -1:
            mask = np.ones(seg_image.shape).astype(np.uint8)
        else:
            mask = np.zeros(seg_image.shape).astype(np.uint8)
            for idx in obj_id:
                mask += (seg_image == idx).astype(np.uint8)
        return mask

    def segment_rgb(self, rgb_image, seg_image, obj_id):
        """
        Mask the rgb image to filter it with rgb for only target object remained

        :param nd.ndarray rgb_image: rendered rgb image
        :param nd.ndarray seg_image: segmentation mask/groundtruth for rendered rgb image
        :param [int] obj_id: list of target object id in simulation

        :return: A masked rgb image with only target object remained
        :rtype: np.ndarray
        """
        rgb_mask = self.extract_mask(seg_image, obj_id)
        rgb_mask = np.expand_dims(rgb_mask, 2)
        rgb_mask = np.repeat(rgb_mask, repeats=3, axis=2)
        rgb_image *= rgb_mask
        rgb_image = o3d.geometry.Image(rgb_image)
        return rgb_image

    def as_o3d_image(self, rgb_image):
        """
        Converts a numpy array image to an open3d Image

        :param nd.ndarray rgb_image: rendered rgb image

        :return: An Open3D rgb image
        :rtype: np.ndarray
        """

        rgb_image = o3d.geometry.Image(rgb_image)
        return rgb_image

    def segment_depth(self, depth_image, seg_image, obj_id):
        """
        Mask the depth image to filter it with only depth for target object remained

        :param nd.ndarray rgb_image: input rgb image from camera image
        :param nd.ndarray seg_image: input segmentation image from camera image
                                     rendered at the same time with rgb image
        :param [int] obj_id: list of target object id in simulation

        :return: A masked rgb image with only target object remained
        :rtype: np.ndarray
        """
        depth_mask = self.extract_mask(seg_image, obj_id)
        depth_image *= depth_mask
        depth_image = o3d.geometry.Image(depth_image)
        return depth_image

    def crop_z(self, pcd, z_crop):
        points = np.asarray(pcd.points)
        normals = np.asarray(pcd.normals)
        colours = np.asarray(pcd.colors)

        pcd_out = o3d.geometry.PointCloud()

        mask = points[:, 2] > z_crop

        points = points[mask, :]
        pcd_out.points = o3d.utility.Vector3dVector(points[:, :3])

        if pcd.has_normals():
            normals = normals[mask, :]
            pcd_out.normals = o3d.utility.Vector3dVector(normals[:, :3])

        if pcd.has_colors():
            colours = colours[mask, :]
            pcd_out.colors = o3d.utility.Vector3dVector(colours[:, :3])

        return pcd_out

    def update_extrinsic(self, view_matrix, eye):
        """
        Update camera extrinsic matrix according to view matrix and eye position

        :param [float] view_matrix:  a list of 16 floats in the format of pybullet view matrix
        :param nd.ndarray eye: 3 element numpy array (x, y, z) as the camera focus position

        :return: Camera position (x, y, z) and orientation in quaternion (x, y, z, w)
        :rtype: np.ndarray
        """
        gl_view_matrix = np.array(view_matrix).reshape(4, 4).T
        rot = R.from_matrix(np.linalg.inv(gl_view_matrix[:3, :3]))
        gl_view_matrix[1, :] *= -1
        gl_view_matrix[2, :] *= -1
        rot = R.from_matrix(np.linalg.inv(gl_view_matrix[:3, :3]))
        quat = rot.as_quat()

        camera_pose = np.eye(4)
        camera_pose[:3, :3] = rot.as_matrix()
        camera_pose[:3, 3] = eye

        self.camera_extrinsic = np.linalg.inv(camera_pose)

        return eye, quat

    def init_volume(self, length=4.0, n_voxel=512, sdf_trunc=0.04):
        """
        Initialize a ScalableTSDFVolume class for volume integration later

        :param float length: length for the TSDF cubic per dimension
        :param int n_voxel: number of voxels per dimension
        :param float sdf_trunc: truncation value for signed distance function (SDF)

        :return: A initialized ScalableTSDFVolume with specified resolution and truncation value
        :rtype: open3d.integration.ScalableTSDFVolume
        """
        volume = o3d.pipelines.integration.ScalableTSDFVolume(
            voxel_length=length / n_voxel,
            sdf_trunc=sdf_trunc,
            color_type=o3d.pipelines.integration.TSDFVolumeColorType.RGB8,
        )
        return volume

    def object_pcd_from_rgbd(
        self, raw_rgb, raw_depth, raw_seg, object_id, vis=False, **kwargs
    ):
        """
        Extract the pointcloud for target object(s) from the rendered rgb-d images and
        segmentation mask

        :param np.ndarray raw_rgb: rendered rgb image
        :param np.ndarray raw_depth: rendered depth image
        :param np.ndarray raw_seg: rendered segmentation groundtruth
        :param [int] object_id: list of target object id from simulation
        :param bool vis: visualization for the extracted pointcloud (for debugging)

        :return: extracted colored pointcloud for target object(s)
        :rtype: open3d.geometry.PointCloud
        """

        self.camera_extrinsic = kwargs.get("extrinsics", np.eye(4))

        if raw_seg is not None:
            rgb_image = self.segment_rgb(raw_rgb, raw_seg, object_id)
            depth_image = self.segment_depth(raw_depth, raw_seg, object_id)
        else:
            rgb_image = self.as_o3d_image(raw_rgb)
            depth_image = self.as_o3d_image(raw_depth)

        rgbd_image = o3d.geometry.RGBDImage.create_from_color_and_depth(
            rgb_image,
            depth_image,
            depth_scale=self.depth_scale,
            depth_trunc=self.depth_truncate,
            convert_rgb_to_intensity=False,
        )
        pcd = o3d.geometry.PointCloud.create_from_rgbd_image(
            rgbd_image, intrinsic=self.camera_intrinsic, extrinsic=self.camera_extrinsic
        )

        if vis:
            visualizePcd(pcd, ratio=0.01, duration=10)

            # import matplotlib.pyplot as plt
            # plt.subplot(1, 2, 1)
            # plt.title('rgb image')
            # plt.imshow(rgbd_image.color)
            # plt.subplot(1, 2, 2)
            # plt.title('depth image')
            # plt.imshow(rgbd_image.depth)
            # plt.show()

            # o3d.visualization.draw_geometries([pcd])

        return pcd

    def object_pcd_from_depth(self, raw_depth, raw_seg, object_id, vis=False, **kwargs):
        """
        Extract the pointcloud for target object(s) from the rendered depth image and
        segmentation mask

        :param np.ndarray raw_depth: rendered depth image
        :param np.ndarray raw_seg: rendered segmentation groundtruth
        :param [int] object_id: list of target object id from simulation
        :param bool vis: visualization for the extracted pointcloud (for debugging)

        :return: extracted colored pointcloud for target object(s)
        :rtype: open3d.geometry.PointCloud
        """

        self.camera_extrinsic = kwargs.get("extrinsics", np.eye(4))
        if raw_seg is not None:
            depth_image = self.segment_depth(raw_depth, raw_seg, object_id)
        else:
            depth_image = self.as_o3d_image(raw_depth)

        pcd = o3d.geometry.PointCloud.create_from_depth_image(
            depth_image,
            intrinsic=self.camera_intrinsic,
            extrinsic=self.camera_extrinsic,
            depth_scale=self.depth_scale,
            depth_trunc=self.depth_truncate,
        )

        if vis:
            visualizePcd(pcd, ratio=0.01, duration=10)
            # o3d.visualization.draw_geometries([pcd])

        return pcd

    def compute_curvature(self, pcd, radius=0.5):
        """
        A slow way to compute the curvature of input pointcloud from
        https://github.com/intel-isl/Open3D/issues/2471

        :param open3d.geometry.PointCloud pcd: input pointcloud
        :param float radius: radius of local region

        :return: list of float value which captures curvature value at each
                 points in the pointcloud
        :rtype: [float]
        """
        points = np.asarray(pcd.points)
        tree = KDTree(points)
        curvature = [0] * points.shape[0]
        for index, point in enumerate(points):
            indices = tree.query_ball_point(point, radius)
            # local covariance
            M = np.array([points[i] for i in indices]).T
            M = np.cov(M)
            # eigen decomposition
            V, E = np.linalg.eig(M)
            # h3 < h2 < h1
            h1, h2, h3 = V
            curvature[index] = h3 / (h1 + h2 + h3)
        return curvature

    def pcd_to_xyzrgba(self, pcd):
        """
        Convert pointcloud from open3d.geometry.PointCloud format into np.ndarray.
        The converted array has the shape as (N, 7) and contain x,y,z, rgba information

        Args:
            pcd (open3d.geometry.PointCloud): input pointcloud

        Returns:
            np.ndarray: (N, 7) array with x, y, z and rgba information of the input pointcloud
        """

        xyz = np.array(pcd.points, dtype=np.float32)
        rgb = np.array(pcd.colors, dtype=np.float32)
        a = np.ones((rgb.shape[0], 1), dtype=rgb.dtype)
        xyzrgba = np.concatenate([xyz, rgb, a], 1)
        return xyzrgba

    def pcd_to_xyzrgbn(self, pcd):
        """
        Convert pointcloud from open3d.geometry.PointCloud format into np.ndarray.
        The converted array has the shape as (N, 7) and contain x,y,z, rgb and normal information

        Args:
            pcd (open3d.geometry.PointCloud): input pointcloud

        Returns:
            np.ndarray: (N, 7) array with x, y, z, rgb, normal information of the input pointcloud
        """
        xyz = np.array(pcd.points, dtype=np.float32)
        rgb = np.array(pcd.colors, dtype=np.float32)
        if not pcd.has_normals():
            pcd.estimate_normals()
        normals = np.array(pcd.normals, dtype=np.float32)
        xyzrgbn = np.concatenate([xyz, rgb, normals], 1)
        return xyzrgbn

    def extract_from_pcd(self, pcd, normal=True, curvature=False):
        """
        Extract target field information (xyz, rgb, normal, curvature) or infer if required from input pointcloud according to specification

        :param open3d.geometry.PointCloud pcd: input pointcloud

        :return: (N, 6/7/8) array with x, y, z, rgb, (optional)normal, (optional) curvature
                 information of the input pointcloud
        :rtype: np.ndarray
        """
        # TODO: make xyz, rgb optional as well?
        xyz = np.array(pcd.points, dtype=np.float32)
        rgb = np.array(pcd.colors, dtype=np.float32)
        if normal:
            if not pcd.has_normals():
                pcd.estimate_normals()
            normals = np.array(pcd.normals, dtype=np.float32)
        else:
            normals = np.zeros(xyz.shape)
        if curvature:
            curvatures = np.array(self.compute_curvature(pcd), dtype=np.float32)
            curvatures = curvatures.reshape(-1, 1)
        else:
            curvatures = np.zeros((xyz.shape[0], 1))
        data = np.concatenate([xyz, rgb], 1)
        if normal:
            data = np.concatenate([data, normals], 1)
        if curvature:
            data = np.concatenate([data, curvatures], 1)
        return data

    # TODO add downsampling script
    def stitch_pcd(self, xyz, rgb, vis=False):
        """
        Stitch pointclouds extracted from multiple views into a single pointcloud.

        :param [np.ndarray] xyz, list of xyz values for pointclouds from multiple views
        :param [np.ndarray] rgb, list of rgb values for pointclouds from multiple views
        :param bool vis: visualize the stitched pointcloud

        :return: The stitched single pointcloud
        :rtype: open3d.geometry.PointCloud
        """
        pcd = o3d.geometry.PointCloud()
        xyz = np.vstack(xyz)
        rgb = np.vstack(rgb)
        pcd.points = o3d.utility.Vector3dVector(xyz)
        pcd.colors = o3d.utility.Vector3dVector(rgb)
        if vis:
            visualizePcd(pcd, ratio=0.001, duration=10)
            # o3d.visualization.draw_geometries([pcd])
        return pcd



#!/usr/bin/env python3
import grip
from grip.environments import PickAndPlaceEnvironment, Env
from grip.agent import DemoAgent
from grip.io import log
import numpy as np
from typing import Dict


class DemoReward(grip.environments.Reward):
    """An example reward function.

    This example computes a 1.0 reward if plate is grasped and placed successfully
    """

    def __init__(self, obj_pos: np.ndarray, obj_ori: np.ndarray):
        """

        Args:
            (obj_pos): initial plate position
            (obj_ori): initial plate orientation
        """

        self.obj_pos0 = obj_pos
        self.obj_ori0 = obj_ori

        self.success = False
        self.was_grasped = False
        self.was_placed = False

    def reward(self, env: Env, obs: Dict) -> float:
        """Computes instantaneous reward given current observation and environment

        Args:
            (env): current task environment
            (obs): current observation from envrironment
        Returns:
            (float): 1.0 reward if plate is grasped and placed successfully, 0.0 otherwise
        """

        height_diff = obs["object"]["pose"][0][2] - self.obj_pos0[2]

        obj_entity = env.get_entity("object")
        robot_entity = env.get_entity("robot")
        table_entity = env.get_entity("table")

        has_grasp_contacts = obj_entity.is_contacting(robot_entity)
        has_table_contacts = obj_entity.is_contacting(table_entity)
        log.info("Has contacts with table {}".format(has_table_contacts))

        if not self.was_grasped and has_grasp_contacts:
            robot_entity.attach_object(robot_entity.tip_link, obj_entity)

        if height_diff < 0.1 and self.was_grasped:
            robot_entity.detach_object(obj_entity)

        log.info("Has contacts with plate {}".format(has_grasp_contacts))

        is_grasped = height_diff >= 0.2 and has_grasp_contacts
        is_placed = not has_grasp_contacts and has_table_contacts

        self.was_grasped = self.was_grasped or is_grasped
        self.was_placed = self.was_placed or is_placed
        log.info(
            "was_grasped {} was_placed {}".format(self.was_grasped, self.was_placed)
        )

        return float(self.was_grasped and self.was_placed)


class EnvRunner:
    """example environment runner: constructs environment and agent and can run one interaction step."""

    def __init__(self, *args, **kwargs):
        self.env = PickAndPlaceEnvironment(seed=None, **kwargs)
        self.agent = DemoAgent(env=self.env)

        self.obs = None
        self.init()

        self.env.world.register_keyhandler(ord("r"), self.init)

    def init(self) -> None:
        """Initialises environment
        Probes initial state, initialises and sets reward
        """

        self.env.step(None)

        self.obs = self.env.reset()
        self.agent.reset()

        obj_pos, obj_ori = self.obs["object"]["pose"]
        self.env.set_reward(DemoReward(obj_pos, obj_ori))

    def run(self) -> None:
        """Runs one agent-environment iteration cycle:
        1. Agent gets an observation
        2. Agent computes an action based on observation
        3. Environment evolves one time step given the agent's action
        """
        self.agent.observe(self.obs)

        action = self.agent.act()

        self.obs, reward, _, _ = self.env.step(action)

        grip.io.log.info("Reward: {}".format(reward))

        grip.sleep(self.env.world.timestep)

    def spin(self):
        """Spins and iterates while connected"""

        while self.env.world.is_connected():
            self.run()


if __name__ == "__main__":
    env_runner = EnvRunner(gripper_max_force=50, enable_texture_randomiser=True)

    env_runner.spin()



from __future__ import annotations
from grip.sensors.camera import RGBDCamera
from ..robot import BulletRobot
from ..io import log
import open3d as o3d
from ..math import position_quaternion2matrix
import numpy as np
from typing import Tuple, Optional
from threading import RLock

import rclpy

from .tf_tools import TFListener, TFPublisher
from .io import xacro_to_urdf_path
import rclpy.node as rclpy_node
import sensor_msgs.msg as sensor_msgs
import cv_bridge
from .ros2_future import wait_for_message
import message_filters
from rclpy.callback_groups import MutuallyExclusiveCallbackGroup


class ROSCamera(RGBDCamera):
    """
    ROSCamera sensor, inherits from RGBDCamera, exposing extra ROS utilities.
    Represents a simulated or a real RGBD camera.

    Args:
        sim (bool, optional): whether or not this camera should be a simulated camera or if it should subscribe to real camera topics. Defaults to True.
        name (str, optional): camera name. Defaults to "CameraSensor"
        position (numpy.ndarray): position of the optical centre of this camera
        orientation (numpy.ndarray): orientation quaternion of the optical centre of this camera
        anchor_robot (grip.robot.BulletRobot, optional): robot that this camera may be anchored/attached to. Defaults to None.
        anchor_link (str, optional): required if anchor_robot is not None. Link name of the anchor_robot that this camera should be attached to.
            Typically a camera link has been already defined in the URDF of the robot, anchor_link can be chosen to be this pre-defined URDF link (camera optical frame)
        cid (int, optional): physics engine unique identifier (e.g. world.id)
        proj_from_intrinsics (bool, optional): if False sets up intrinsics, sets up projection matrix using self.compute_projection_matrix
            otherwise sets projection matrix using self.build_projection_matrix. Defaults to True.
        base_frame_id (str, optional): name of the base frame id. Default: "base_link".
        broadcast_tf_on (bool, optional): Applicable if sim=True. If broadcast_tf_on is also set to True, then whenever self.obs() is called it then broadcasts the transform tree of this camera. Link names that do not contain this camera name as a substring are ignored (not published). This flag applicable in simulation only.
        parent_node (rclpy_node.Node, optional): parent ROS2 node to which this ROSCamera may belong to. If not passed, this ROSCamera will be its own node. Defaults to None meaning this camera will have its own ROS2 node.

    """

    def __init__(self, **kwargs):
        self.sim = kwargs.get("sim", True)
        self.name = kwargs.get("name", "camera")
        self.broadcast_tf_on = kwargs.get("broadcast_tf_on", False)
        self.base_frame_id = kwargs.get("base_frame_id", "base_link")
        self.node = kwargs.get("parent_node", rclpy_node.Node(f"{self.name}_node"))

        self.image_topic = kwargs.get(
            "image_topic", f"{self.name}/color/undistorted/image_rect"
        )
        self.image_info_topic = kwargs.get(
            "image_info_topic", f"{self.name}/color/undistorted/camera_info"
        )

        self.depth_topic = kwargs.get(
            "depth_topic", f"{self.name}/depth_registered/undistorted/image_rect"
        )

        self.depth_info_topic = kwargs.get(
            "depth_info_topic", f"{self.name}/depth_registered/undistorted/camera_info"
        )

        kwargs.setdefault("width", 424)
        kwargs.setdefault("height", 280)
        kwargs.setdefault("camera_far", 4.0)
        kwargs.setdefault("proj_from_intrinsics", False)

        self.camera_info_msg = sensor_msgs.CameraInfo()

        self._timer_cb_group = MutuallyExclusiveCallbackGroup()

        self.depth_msg = None
        self.image_msg = None

        super().__init__(**kwargs)

        self.is_ready = True
        self.image_info_ready = True

        if self.sim:
            self._setup_sim(**kwargs)
        else:
            self._setup_real()

        self.rgb_image = None
        self.depth_image = None
        self.seg_image = None

        self.image_bridge = cv_bridge.CvBridge()

        self.mutex = RLock()

        self._cb_lock = kwargs.get("lock", RLock())
        self._timer_func = None
        self._async_on = False

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        """
        Shuts down this camera
        """

        if not hasattr(self, "_cb_lock"):
            return

        if self.sim:
            with self._cb_lock:
                self.stop_timer_publisher()
                for pub in [
                    self.pub_image,
                    self.pub_image_info,
                    self.pub_depth,
                    self.pub_depth_info,
                ]:
                    try:
                        pub.destroy()
                    except Exception as e:
                        log.warning(
                            f"Could not kill publisher {pub.topic_name}. Error: {e}"
                        )

    def destroy(self) -> None:
        """Destroy underlying ROS handles"""

        self.shutdown()
        self.node.destroy_node()
        self.anchor_robot = None

    def _init_camera_info(self) -> bool:
        """
        Initialises camera info for the real camera (intrinsics), if self.sim = False.

        Returns:
            bool: whether or not received camera info successfully
        """

        success = False
        try:
            success, self.camera_info_msg = wait_for_message(
                sensor_msgs.CameraInfo,
                self.node,
                self.depth_info_topic,
                time_to_wait=5,
            )

            if success:
                log.info("Camera info received successfully!")
            else:
                log.warning("Camera info has not been received.")
        except Exception as e:
            log.warning(
                f"Waiting for {self.name} info timeout. Is the camera on? Initialising camera info with empty camera info. Error: {e}"
            )

        return success

    def _setup_sim(self, **kwargs) -> None:
        self.pub_image = self.node.create_publisher(
            sensor_msgs.Image, self.image_topic, 10
        )

        self.pub_image_info = self.node.create_publisher(
            sensor_msgs.CameraInfo, self.image_info_topic, 10
        )

        self.pub_depth = self.node.create_publisher(
            sensor_msgs.Image, self.depth_topic, 10
        )

        self.pub_depth_info = self.node.create_publisher(
            sensor_msgs.CameraInfo, self.depth_info_topic, 10
        )

        self.image_msg = None
        self.depth_msg = None
        self.camera_info_msg = None

        self.image_msg = sensor_msgs.Image()
        self.depth_msg = sensor_msgs.Image()

        self.image_msg.width = self.width
        self.image_msg.height = self.height
        self.image_msg.encoding = kwargs.get("colour_encoding", "rgb8")
        self.image_msg.is_bigendian = 0
        self.image_msg.step = self.image_msg.width * 3 * 1

        self.depth_msg.width = self.image_msg.width
        self.depth_msg.height = self.image_msg.height
        self.depth_msg.encoding = kwargs.get("depth_encoding", "32FC1")
        self.depth_msg.is_bigendian = self.image_msg.is_bigendian
        self.depth_msg.step = self.depth_msg.width * 1 * 4  # Width*Channels*nBytes

        camera_info = self.get_camera_info()

        self.camera_info_msg = sensor_msgs.CameraInfo(**camera_info)

        self.image_msg.header.frame_id = self.frame_id
        self.depth_msg.header.frame_id = self.frame_id
        self.camera_info_msg.header.frame_id = self.frame_id

        self.tf_publisher = TFPublisher(self.node)

    def _setup_real(self) -> None:
        """
        Sets up real camera. Registers to depth and rgb topics.
        """

        # from rclpy.qos import QoSProfile, QoSReliabilityPolicy, QoSHistoryPolicy
        # qos_profile = QoSProfile(
        #     reliability=QoSReliabilityPolicy.RMW_QOS_POLICY_RELIABILITY_BEST_EFFORT#RMW_QOS_POLICY_RELIABILITY_BEST_EFFORT,
        #     history=QoSHistoryPolicy.RMW_QOS_POLICY_HISTORY_KEEP_LAST,
        #     depth=1
        # )

        self.is_ready = False
        self.image_info_ready = self._init_camera_info()

        self.tf_listener = TFListener(self.node)

        log.debug(f"Creating filter for topic: {self.depth_topic}")
        self.depth_sub = message_filters.Subscriber(
            self.node,
            sensor_msgs.Image,
            self.depth_topic,
        )

        log.debug(f"Creating filter for topic: {self.image_topic}")
        self.rgb_sub = message_filters.Subscriber(
            self.node,
            sensor_msgs.Image,
            self.image_topic,
        )

        self.time_synchroniser = message_filters.ApproximateTimeSynchronizer(
            [self.depth_sub, self.rgb_sub], queue_size=1, slop=1e9
        )

        self.time_synchroniser.registerCallback(self._subscriber_callback)

    # overriden
    def obs(
        self, ret_ext: bool = False
    ) -> Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]:
        """
        Retrieves an observation from this camera.

        Args:
            ret_ext (bool, optional): whether of not it should return the extrinsics parameters of the camera as well.
        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, Optional[numpy.ndarray]]: three ndarrays rgb image shape-(width, height), depth image shape-(width, height),
                the third array (optional return) being the extrinsics of the camera with shape-(4,4) if ret_ext=True.
        """
        if self.sim:
            return self._obs_sim(ret_ext=ret_ext)

        return self._obs_real(ret_ext=ret_ext)

    def get_cloud_obs(self, full_ret=False) -> Tuple[
        o3d.geometry.PointCloud,
        Optional[o3d.geometry.RGBDImage],
        Optional[o3d.camera.PinholeCameraIntrinsic],
        Optional[np.ndarray],
    ]:
        """
        Gets open3d cloud observation.

        Args:
            full_ret (bool, optional): If full_ret=False, returns only point cloud observartion. If full_ret=True returns pcd, rgbd image, intrinsics, extrinsics.

        Returns:
            Tuple[o3d.geometry.PointCloud, Optional[o3d.geometry.RGBDImage], Optional[o3d.camera.PinholeCameraIntrinsic], Optional[np.ndarray]]: Point cloud observation and optionally rgbd_image, camera intrinsics and extrinsics if full_ret=True
        """

        colour, depth, extrinsics = self.obs(ret_ext=True)

        if colour is None or depth is None or extrinsics is None:
            return None

        pose = np.linalg.inv(extrinsics)

        intrinsics = self.get_o3d_intrinsics()

        log.debug(f"Intrinsics: {intrinsics.intrinsic_matrix}")

        rgb_image = self.pcd_parser.as_o3d_image(colour)
        depth_image = self.pcd_parser.as_o3d_image(depth)

        rgbd_image = o3d.geometry.RGBDImage.create_from_color_and_depth(
            rgb_image,
            depth_image,
            depth_scale=1,
            depth_trunc=self.pcd_parser.depth_truncate,
            convert_rgb_to_intensity=False,
        )

        pcd = o3d.geometry.PointCloud.create_from_rgbd_image(
            rgbd_image,
            intrinsic=intrinsics,
            extrinsic=extrinsics,
            project_valid_depth_only=False,
        )

        log.debug(f"Point cloud size: {len(pcd.points)}")
        pcd.estimate_normals()
        pcd.orient_normals_towards_camera_location(camera_location=pose[:3, 3])

        ret = pcd

        if full_ret:
            ret = (pcd, rgbd_image, intrinsics, extrinsics)

        return ret

    def _obs_real(
        self, ret_ext: bool = False
    ) -> Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]:
        """
        Retrieves an observation from this (real) camera.

        Args:
            ret_ext (bool, optional): whether of not it should return the extrinsics parameters of the camera as well.
        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, Optional[numpy.ndarray]]: three ndarrays rgb image shape-(width, height), depth image shape-(width, height),
                the third array (optional return) being the extrinsics of the camera with shape-(4,4) if ret_ext=True.
        """

        rgb, depth, extrinsics = None, None, None

        with self.mutex:
            if self.ready:
                rgb, depth = self.rgb_image, self.depth_image
                extrinsics = self.get_extrinsics(self.depth_msg.header.stamp)
            else:
                log.warning("Camera is not ready yet.")
                raise RuntimeWarning("Camera is not ready yet.")

            if ret_ext:
                return rgb, depth, extrinsics

            return rgb, depth

    def _obs_sim(
        self, ret_ext: bool = False
    ) -> Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]:
        """
        Retrieves an observation from this (simulated) camera.

        Args:
            ret_ext (bool, optional): whether of not it should return the extrinsics parameters of the camera as well.
        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, Optional[numpy.ndarray]]: three ndarrays rgb image shape-(width, height), depth image shape-(width, height),
                the third array (optional return) being the extrinsics of the camera with shape-(4,4) if ret_ext=True.
        """
        self.rgb_image, self.depth_image, self.seg_image = super().obs()

        self.image_msg.data = self.image_bridge.cv2_to_imgmsg(
            self.rgb_image, encoding=self.image_msg.encoding
        ).data

        self.depth_msg.data = self.image_bridge.cv2_to_imgmsg(
            self.depth_image, encoding=self.depth_msg.encoding
        ).data

        stamp = self.node.get_clock().now().to_msg()
        self.image_msg.header.stamp = stamp
        self.depth_msg.header.stamp = stamp
        self.camera_info_msg.header.stamp = stamp

        self.pub_image.publish(self.image_msg)
        self.pub_image_info.publish(self.camera_info_msg)

        self.pub_depth.publish(self.depth_msg)
        self.pub_depth_info.publish(self.camera_info_msg)

        if self.sim and self.broadcast_tf_on and self.anchor_robot:
            self._broadcast_anchor_tfs(stamp)

        self.tf_publisher.broadcast_frame(
            self.pose, self.base_frame_id, self.frame_id, stamp
        )

        if ret_ext:
            extrinsics = self.get_extrinsics(self.depth_msg.header.stamp)
            return self.rgb_image, self.depth_image, extrinsics

        return self.rgb_image, self.depth_image

    def _broadcast_anchor_tfs(self, stamp: rclpy.time.Time) -> None:
        """
        Broadcasts tf tree of anchored robot (kinematic chain)

        Args:
            stamp: time stamp.
        """

        for link_name in self.anchor_robot.get_link_names():
            if self.name in link_name:
                pose = self.anchor_robot.get_link_pose_by_name(link_name)
                self.tf_publisher.broadcast_frame(
                    pose, link_name, self.base_frame_id, stamp
                )

    def _publisher_callback(self) -> None:
        with self._cb_lock:
            try:
                self.obs()
            except Exception as e:
                log.warning(
                    f"Shutting down {self.name}. Failed to query camera observation. Perhaps disconnected from world. Message {e}"
                )

    def setup_timer_publisher(self, fps: int = 10) -> None:
        """
        Sets up and starts a timer callback. Image topics will be published in the background for asynchronous usage.
        """
        if self._async_on:
            return

        self._timer_func = self.node.create_timer(
            1.0 / fps, self._publisher_callback, callback_group=self._timer_cb_group
        )

        self._async_on = True

    def stop_timer_publisher(self) -> None:
        """
        Stops timer publisher. Disables asyncronous publishing.
        """

        with self._cb_lock:
            if not self._async_on:
                return

            self._timer_func.destroy()

        self._timer_func = None
        self._async_on = False

    def _get_intrinsics_real(self) -> Tuple[float, float, float, float, float]:
        info_msg = self.camera_info_msg

        fx = info_msg.k[0]
        fy = info_msg.k[4]

        cx = info_msg.k[2]
        cy = info_msg.k[5]

        width = info_msg.width
        height = info_msg.height

        return width, height, fx, fy, cx, cy

    # override
    def get_intrinsics(self) -> np.ndarray:
        """
        Gets the raw intrinsics parameters of this camera.

        Returns:
            Tuple[float, float, float, float, float, float]: intrinsics parameters width, height, fx, fy, cx, cy
        """

        if self.sim:
            return super().get_intrinsics()

        return self._get_intrinsics_real()

    def get_intrinsic_matrix(self) -> np.ndarray:
        """
        Gets the camera intrinsic parameters of this camera as a matrix.

        Returns:
            (numpy.ndarray): shape-(3,3) matrix of intrinsic parameters of this camera
        """

        _, _, fx, fy, cx, cy = self.get_intrinsics()

        intrinsic_matrix = np.array([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])

        return intrinsic_matrix

    def get_segmentation(self) -> np.ndarray:
        """
        Gets the ground-truth segmentation image/mask.


        Returns:
            (numpy.ndarray): shape-(width, height) segmentation mask image (ground-truth segmentation).
        """

        _, _, self.seg_image = super().obs()

        return self.seg_image

    def _get_real_extrinsics(self, time: rclpy.time.Time = None) -> np.ndarray:
        position, orientation, time = self.tf_listener.get_transform(
            self.camera_info_msg.header.frame_id, self.base_frame_id, time
        )

        if position is None or orientation is None:
            camera_extrinsics = None
        else:
            camera_extrinsics = np.array(
                position_quaternion2matrix(position, orientation)
            )

        return camera_extrinsics

    # override
    def get_extrinsics(self, time: rclpy.time.Time = rclpy.time.Time()) -> np.ndarray:
        """
        Gets the exintrics matrix of this camera (the inverse of its pose).

        Args:
            time (rclpy.time.Time, optional): gets extrinsics at specified time. Defaults to rclpy.time.Time(), meaning most current measurement.
        Returns:
            (numpy.ndarray): camera extrinsics shape-(4,4)
        """

        ret = None
        if self.sim:
            extrinsics = super().get_extrinsics()

            if self.anchor_robot:
                base_frame = self.anchor_robot.get_link_pose_by_name(self.base_frame_id)
                base_frame = position_quaternion2matrix(*base_frame)

                ret = extrinsics @ base_frame
            else:
                ret = extrinsics
        else:
            ret = self._get_real_extrinsics(time)

        return ret

    def get_transform(self) -> np.ndarray:
        """
        Gets the transformation matrix representing the pose of this camera.
        Returns the same as self.pose, but in matrix form.

        Returns:
            (numpy.ndarray): camera pose transform shape-(4,4)
        """
        return position_quaternion2matrix(*self.pose)

    def _get_real_pose(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns real camera pose.

        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): camera pose, position and quaternion
        """
        position, orientation, _ = self.tf_listener.get_transform(
            self.base_frame_id, self.camera_info_msg.header.frame_id, rclpy.time.Time()
        )
        return position, orientation

    # override
    @property
    def pose(self):
        """
        (Tuple[numpy.ndarray, numpy.ndarray]): camera pose, position and quaternion
        """
        if self.sim:
            return self.position, self.orientation
        else:
            return self._get_real_pose()

    def _subscriber_callback(self, depth_msg, rgb_msg) -> None:
        with self.mutex:
            self.image_msg = rgb_msg
            self.depth_msg = depth_msg

            self.rgb_image = None
            try:
                self.rgb_image = self.image_bridge.imgmsg_to_cv2(rgb_msg, "rgb8")

                match self.depth_msg.encoding:
                    case "16UC1":
                        self.depth_image = np.frombuffer(
                            depth_msg.data, dtype=np.uint16
                        ).reshape(depth_msg.height, depth_msg.width)
                        self.depth_image = np.multiply(
                            self.depth_image.astype("float32"), 0.001
                        )
                    case _:
                        self.depth_image = self.image_bridge.imgmsg_to_cv2(
                            depth_msg, desired_encoding="32FC1"
                        )

                # this camera is ready if messages have been received successfully including image info
                self.is_ready = self.image_info_ready

                # import cv2
                # cv2.imshow("o3d image", self.rgb_image)

                # cv2.imshow("o3d depth", self.depth_image)

                # print("Depth: ", np.median(self.depth_image))
                # cv2.waitKey(1)

                log.debug("Received camera images")

            except cv_bridge.CvBridgeError as e:
                log.error(f"Error on converting image to rgb8 format {e}")

    def ready(self) -> bool:
        """
        Indicates whether or not this camera is ready.

        Returns:
            (bool): whether or not this camera is ready.
        """
        return self.is_ready

    @classmethod
    def make_d415(
        cls,
        camera_name: str,
        position: np.ndarray,
        orientation: np.ndarray,
        sim: bool,
        lock: RLock = None,
        auto_anchor: bool = False,
        phys_id: int = 0,
    ) -> ROSCamera:
        """

        Creates a grip.sensors.ROSCamera using the intel d415 realsense urdf description.

        Args:
            camera_name: the name of the camera.
            position: the position of the camera optical centre with respect to world frame.
            orientation: the orientation of the camera optical with respect to world frame.
            sim: whether or not this is a simulated camera or not. If sim=False it will try to connect to a real camera publishing in standard topics.
            lock (optional): camera creates an internal lock to deal with critical memory regions. If global thread coordination is needed a global lock can be passed so this camera can use it instead.
            auto_anchor (optional): whether or not this camera should create its own rigid body chain in pybullet and anchor itself to it or not.
            phys_id (optional): physics id where camera is going be inserted to.

        Returns:
            (ROSCamera): the created camera.



        """
        camera_optical_link_frame = None
        camera_chain = None

        if auto_anchor:
            camera_optical_link_frame = f"{camera_name}_color_optical_frame"

            urdf_file = xacro_to_urdf_path(
                "sensor_descriptions",
                "intel_ds415_camera_instance.urdf.xacro",
                "intel_ds415_camera.urdf",
                xacro_args=f"name:='{camera_name}'",
                pkg_subfolder="intel_ds415",
            )

            camera_chain = BulletRobot(
                urdf_file=urdf_file,
                use_fixed_base=True,
                has_kinematics=False,
                phys_id=phys_id,
            )

            camera_chain.disable_collision()
            camera_chain.set_link_pose(
                camera_optical_link_frame,
                position,
                orientation,
            )

        camera = cls(
            cid=phys_id,
            name=camera_name,
            position=position,
            orientation=orientation,
            anchor_robot=camera_chain,
            anchor_link=camera_optical_link_frame,
            sim=sim,
            broadcast_tf_on=True,
            lock=lock,
        )

        return camera



from scipy.interpolate import CubicSpline
import numpy as np
from .minjerk import (
    minjerk_coefficients,
    minjerk_trajectory,
)
from ..io import import_module, log


class Waypoint(object):
    def __init__(self, positions, velocities, accelearations):
        self.positions = positions
        self.velocities = velocities
        self.accelearations = accelearations


class Path(object):
    def __init__(self, **kwargs):
        self.points = kwargs.get("points", None)
        self.velocities = kwargs.get("velocities", None)
        self.accelerations = kwargs.get("accelerations", None)
        self.times = kwargs.get("times", None)
        self.joint_names = kwargs.get("joint_names", [])
        self._path = None
        self._cspline = None

    @property
    def path(self):
        return self._path

    @property
    def size(self):
        return len(self.points)

    def __len__(self):
        return self.size

    @property
    def curve_length(self):
        s = np.linspace(0.0, 1.0, len(self.points))
        ys = self.cubic_spline(s, 1)

        length = 0

        # Euler integration
        dt = 1.0 / float(len(self.points))

        for y in ys:
            length += np.linalg.norm(y) * dt

        return length

    def reversed(self):
        self.points = np.flip(self.points, 0)
        # self.velocities = None if self.velocities is None else np.flip(self.velocities,0)
        # self.accelerations = None if self.accelerations is None else np.flip(self.accelerations,0)

        self._cspline = None

    def reverse(self):
        points = None if self.points is None else np.flip(self.points, 0)
        # velocities = None if self.velocities is None else np.flip(self.velocities,0)
        # accelerations = None if self.accelerations is None else np.flip(self.accelerations,0)

        return Path(
            points=points,
            # velocities=velocities,
            # accelerations=accelerations,
            times=self.times,
            joint_names=self.joint_names,
        )

    def cubic_spline(self, xs, derivative_order=0, extrapolate=True, use_cache=True):
        if self._cspline is None or not use_cache:
            s = np.linspace(0.0, 1.0, len(self.points))
            self._cspline = CubicSpline(s, self.points, extrapolate=extrapolate)

        return self._cspline(xs, derivative_order)

    def cubic_spline_path(
        self, xs, derivative_order=0, extrapolate=True, use_cache=True
    ):
        points = self.cubic_spline(
            xs,
            derivative_order=derivative_order,
            extrapolate=extrapolate,
            use_cache=use_cache,
        )

        return Path(points=points)

    def concat(self, other, ai=0, bi=0):
        return Path(points=np.vstack([self.points[ai:, :3], other.points[bi:, :3]]))

    def _join_array(self, arr0, arr1):
        arr = None
        if arr0 is not None and arr1 is not None:
            arr = np.vstack([arr0, arr1])

        return arr

    def join(self, other):
        points = self._join_array(self.points, other.points)
        velocities = self._join_array(self.velocities, other.velocities)
        accelerations = self._join_array(self.accelerations, other.accelerations)

        times = None
        if self.times is not None and other.times is not None:
            times = np.hstack([self.times, other.times + self.times[-1]])

        return Path(
            points=points,
            velocities=velocities,
            accelerations=accelerations,
            times=times,
            joint_names=self.joint_names,
        )

    def order(self, anchor_point):
        dists = [np.linalg.norm(p - anchor_point) for p in self.points]

        sidx = np.argsort(dists)[::-1]

        sorted_points = self.points[sidx, :]

        self._cspline = None

        self.points = sorted_points

    def time_parametrise(self, total_time=10, dt=0.2):
        npts = np.ceil(total_time / dt)
        ts = np.linspace(0.0, 1.0, int(npts))
        points = self.cubic_spline(ts, extrapolate=False, use_cache=False)
        velocities = self.cubic_spline(ts, derivative_order=1, extrapolate=False)
        accelerations = self.cubic_spline(ts, derivative_order=2, extrapolate=False)

        return Path(
            points=points,
            velocities=velocities,
            accelerations=accelerations,
            times=ts * total_time,
            joint_names=self.joint_names,
        )

    def __iter__(self):
        return iter(self.points)

    def __getitem__(self, index):
        positions = self.points[index, :]
        velocities = None if self.velocities is None else self.velocities[index, :]
        accelerations = (
            None if self.accelerations is None else self.accelerations[index, :]
        )

        return Waypoint(positions, velocities, accelerations)

    def compute_mjt(self, num_intervals=5):
        if self.points is None or self.times is None:
            return

        point_durations = [
            self.times[i + 1] - self.times[i] for i in range(len(self.times) - 1)
        ]
        # n_joints = len(self.points[0])

        # dimensions_dict = {'positions':True,
        #                     'velocities':False,
        #                     'accelerations':False}
        # self.m_matrix = compute_minjerk_coeff(n_joints,
        #                                  self.points,
        #                                  point_durations,
        #                                  dimensions_dict)

        # ts = np.linspace(0.0, 1.0, len(self.times))
        m_coeffs = minjerk_coefficients(self.points)
        points = np.array(minjerk_trajectory(m_coeffs, num_intervals, point_durations))
        times = np.linspace(0.0, self.times[-1], len(points))

        self.points = points
        self.times = times

    def as_ros_trajectory(self):
        tj = import_module("trajectory_msgs.msg")
        rclpy = import_module("rclpy")

        if rclpy is None:
            log.warning("rclpy is not available.")
            return
        if tj is None:
            log.warning("trajectory_msgs.msg is not available.")
            return

        trajectory = tj.JointTrajectory()
        trajectory.joint_names = self.joint_names

        for i in range(self.size):
            point = tj.JointTrajectoryPoint()

            point.positions = self.points[i, :].tolist()

            if self.velocities is not None:
                point.velocities = self.velocities[i, :].tolist()
            if self.accelerations is not None:
                point.accelerations = self.accelerations[i, :].tolist()

            point.time_from_start = rclpy.duration.Duration(
                seconds=self.times[i]
            ).to_msg()

            trajectory.points.append(point)

        return trajectory

    def as_ros_trajectory_action_goal(self, goal_time_tolerance=1.0):
        cm = import_module("control_msgs.action")
        rclpy = import_module("rclpy")

        trajgoal = cm.FollowJointTrajectory.Goal()

        trajgoal.trajectory = self.as_ros_trajectory()
        trajgoal.goal_time_tolerance = rclpy.duration.Duration(
            seconds=goal_time_tolerance
        ).to_msg()  # seconds

        return trajgoal

    def plot(self):
        import matplotlib.pyplot as plt

        for arr in [self.points, self.velocities, self.accelerations, self.times]:
            plt.figure()
            plt.plot(arr)

        plt.show()

    @classmethod
    def from_ros_trajectory_action_goal(cls, goal):
        points = np.asarray([np.asarray(p.positions) for p in goal.trajectory.points])
        velocities = np.asarray(
            [np.asarray(p.velocities) for p in goal.trajectory.points]
        )
        accelerations = np.asarray(
            [np.asarray(p.accelerations) for p in goal.trajectory.points]
        )

        times = np.asarray(
            [
                (p.time_from_start.sec + p.time_from_start.nanosec * 1e-9)
                for p in goal.trajectory.points
            ]
        )
        return cls(
            points=points,
            velocities=velocities,
            accelerations=accelerations,
            times=times,
            joint_names=goal.trajectory.joint_names,
        )

    @classmethod
    def from_plex_traj(cls, trajectory):
        return cls(
            points=np.array(trajectory.positions),
            velocities=np.array(trajectory.velocities),
            accelerations=np.array(trajectory.accelerations),
            times=np.array(trajectory.times),
            joint_names=trajectory.jointNames,
        )



"""
Serialization of sensor_msgs.PointCloud2 messages. Many of these are copied from sensor_msgs.point_cloud2.py in ros1.
"""

#! /usr/bin/python3
import ctypes
import math
import struct
import copy

import numpy as np
import open3d as o3d
from typing import List

# from nptyping import NDArray, Float32, Int32, Shape
from enum import Enum, unique

from std_msgs.msg import Header
from geometry_msgs.msg import Pose, Point, Quaternion
from sensor_msgs.msg import PointCloud2, PointField


@unique
class PointCloudType(Enum):
    """
    Enum defining the type of point cloud, where type is determined by the data present.
    """

    XYZ = 0
    XYZ_RGB = 1
    XYZ_NORMAL = 2
    XYZ_NORMAL_RGB = 3
    XYZ_NORMAL_CURV_RGB = 4

    @classmethod
    def __missing__(cls, value):
        """
        If no value or invalid value is specified, defaults to XYZ
        """
        return cls.XYZ


_DATATYPES = {}
_DATATYPES[PointField.INT8] = ("b", 1)
_DATATYPES[PointField.UINT8] = ("B", 1)
_DATATYPES[PointField.INT16] = ("h", 2)
_DATATYPES[PointField.UINT16] = ("H", 2)
_DATATYPES[PointField.INT32] = ("i", 4)
_DATATYPES[PointField.UINT32] = ("I", 4)
_DATATYPES[PointField.FLOAT32] = ("f", 4)
_DATATYPES[PointField.FLOAT64] = ("d", 8)


FIELDS_ORDER = ["x", "y", "z", "normal_x", "normal_y", "normal_z", "rgb", "curvature"]


def read_points(cloud, field_names=None, skip_nans=True, uvs=None):
    """
    Read points from a L{sensor_msgs.PointCloud2} message.
    @param cloud: The point cloud to read from.
    @type  cloud: L{se_nsor_msgs.PointCloud2}
    @param field_names: The names of fields to read. If None, read all fields. [default: None]
    @type  field_names: iterable
    @param skip_nans: If True, then don't return any point with a NaN value.
    @type  skip_nans: bool [default: False]
    @param uvs: If specified, then only return the points at the given coordinates. [default: empty list]
    @type  uvs: iterable
    @return: Generator which yields a list of values for each point.
    @rtype:  generator
    """
    assert isinstance(cloud, PointCloud2), "cloud is not a sensor_msgs.msg.PointCloud2"
    fmt = _get_struct_fmt(cloud.is_bigendian, cloud.fields, field_names)
    width, height, point_step, row_step, data, isnan = (
        cloud.width,
        cloud.height,
        cloud.point_step,
        cloud.row_step,
        cloud.data,
        math.isnan,
    )
    unpack_from = struct.Struct(fmt).unpack_from

    if skip_nans:
        if uvs:
            for u, v in uvs:
                p = unpack_from(data, (row_step * v) + (point_step * u))
                has_nan = False
                for pv in p:
                    if isnan(pv):
                        has_nan = True
                        break
                if not has_nan:
                    yield p
        else:
            for v in range(height):
                offset = row_step * v
                for u in range(width):
                    p = unpack_from(data, offset)
                    has_nan = False
                    for pv in p:
                        if isnan(pv):
                            has_nan = True
                            break
                    if not has_nan:
                        yield p
                    offset += point_step
    else:
        if uvs:
            for u, v in uvs:
                yield unpack_from(data, (row_step * v) + (point_step * u))
        else:
            for v in range(height):
                offset = row_step * v
                for u in range(width):
                    yield unpack_from(data, offset)
                    offset += point_step


def create_cloud(header, fields, points):
    """
    Create a L{sensor_msgs.msg.PointCloud2} message.
    @param header: The point cloud header.
    @type  header: L{std_msgs.msg.Header}
    @param fields: The point cloud fields.
    @type  fields: iterable of L{sensor_msgs.msg.PointField}
    @param points: The point cloud points.
    @type  points: list of iterables, i.e. one iterable for each point, with the
                   elements of each iterable being the values of the fields for
                   that point (in the same order as the fields parameter)
    @return: The point cloud.
    @rtype:  L{sensor_msgs.msg.PointCloud2}
    """

    cloud_struct = struct.Struct(_get_struct_fmt(False, fields))

    buff = ctypes.create_string_buffer(cloud_struct.size * len(points))

    point_step, pack_into = cloud_struct.size, cloud_struct.pack_into
    offset = 0
    i = 0
    has_colour = False
    while i < len(fields):
        if fields[i].name == "rgb":
            has_colour = True
            break
        i += 1

    for p in points:
        values = [v for v in p]

        if has_colour:
            values[i] = int(values[i])

        pack_into(buff, offset, *values)
        offset += point_step

    return PointCloud2(
        header=header,
        height=1,
        width=len(points),
        is_dense=False,
        is_bigendian=False,
        fields=fields,
        point_step=cloud_struct.size,
        row_step=cloud_struct.size * len(points),
        data=buff.raw,
    )


def create_cloud_xyz32(header, points):
    """
    Create a L{sensor_msgs.msg.PointCloud2} message with 3 float32 fields (x, y, z).
    @param header: The point cloud header.
    @type  header: L{std_msgs.msg.Header}
    @param points: The point cloud points.
    @type  points: iterable
    @return: The point cloud.
    @rtype:  L{sensor_msgs.msg.PointCloud2}
    """
    fields = [
        PointField(name="x", offset=0, datatype=PointField.FLOAT32, count=1),
        PointField(name="y", offset=4, datatype=PointField.FLOAT32, count=1),
        PointField(name="z", offset=8, datatype=PointField.FLOAT32, count=1),
    ]
    return create_cloud(header, fields, points)


def create_cloud_xyz32color(header, points):
    """
    Create a L{sensor_msgs.msg.PointCloud2} message with 3 float32 fields (x, y, z).
    @param header: The point cloud header.
    @type  header: L{std_msgs.msg.Header}
    @param points: The point cloud points.
    @type  points: iterable
    @return: The point cloud.
    @rtype:  L{sensor_msgs.msg.PointCloud2}
    """

    fields = [
        PointField(name="x", offset=0, datatype=PointField.FLOAT32, count=1),
        PointField(name="y", offset=4, datatype=PointField.FLOAT32, count=1),
        PointField(name="z", offset=8, datatype=PointField.FLOAT32, count=1),
        PointField(name="rgb", offset=12, datatype=PointField.UINT32, count=1),
    ]
    return create_cloud(header, fields, points)


def create_cloud_point_normal_color_curvature(header, points):
    """
    Create a L{sensor_msgs.msg.PointCloud2} message with 3 float32 fields (x, y, z).
    @param header: The point cloud header.
    @type  header: L{std_msgs.msg.Header}
    @param points: The point cloud points.
    @type  points: iterable
    @return: The point cloud.
    @rtype:  L{sensor_msgs.msg.PointCloud2}
    """

    # ['x', 'y', 'z', 'normal_x', 'normal_y', 'normal_z', 'rgb', 'curvature', 'principal_curvature_x', 'principal_curvature_y', 'principal_curvature_z', 'pc1', 'pc2']
    fields = [
        PointField(name="x", offset=0, datatype=PointField.FLOAT32, count=1),
        PointField(name="y", offset=4, datatype=PointField.FLOAT32, count=1),
        PointField(name="z", offset=8, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_x", offset=12, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_y", offset=16, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_z", offset=20, datatype=PointField.FLOAT32, count=1),
        PointField(name="curvature", offset=24, datatype=PointField.FLOAT32, count=1),
        PointField(name="rgb", offset=28, datatype=PointField.UINT32, count=1),
        #   PointField('curvature', 28, PointField.FLOAT32, 1),
        #   PointField('principal_curvature_x', 32, PointField.FLOAT32, 1),
        #   PointField('principal_curvature_y', 36, PointField.FLOAT32, 1),
        #   PointField('principal_curvature_z', 40, PointField.FLOAT32, 1),
        #   PointField('pc1', 42, PointField.FLOAT32, 1),
        #   PointField('pc2', 46, PointField.FLOAT32, 1)
    ]
    return create_cloud(header, fields, points)


def create_cloud_point_color_normal(header, points):
    """
    Create a L{sensor_msgs.msg.PointCloud2} message with 3 float32 fields (x, y, z).
    @param header: The point cloud header.
    @type  header: L{std_msgs.msg.Header}
    @param points: The point cloud points.
    @type  points: iterable
    @return: The point cloud.
    @rtype:  L{sensor_msgs.msg.PointCloud2}
    """

    # ['x', 'y', 'z', 'normal_x', 'normal_y', 'normal_z', 'rgb']
    fields = [
        PointField(name="x", offset=0, datatype=PointField.FLOAT32, count=1),
        PointField(name="y", offset=4, datatype=PointField.FLOAT32, count=1),
        PointField(name="z", offset=8, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_x", offset=12, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_y", offset=16, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_z", offset=20, datatype=PointField.FLOAT32, count=1),
        PointField(name="rgb", offset=24, datatype=PointField.UINT32, count=1),
    ]
    return create_cloud(header, fields, points)


def create_cloud_point_normal(header, points):
    """
    Create a L{sensor_msgs.msg.PointCloud2} message with 6 float32 fields (x, y, z, normal_x, normal_y, normal_z).
    @param header: The point cloud header.
    @type  header: L{std_msgs.msg.Header}
    @param points: The point cloud points.
    @type  points: iterable
    @return: The point cloud.
    @rtype:  L{sensor_msgs.msg.PointCloud2}
    """

    # ['x', 'y', 'z', 'normal_x', 'normal_y', 'normal_z']
    fields = [
        PointField(name="x", offset=0, datatype=PointField.FLOAT32, count=1),
        PointField(name="y", offset=4, datatype=PointField.FLOAT32, count=1),
        PointField(name="z", offset=8, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_x", offset=12, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_y", offset=16, datatype=PointField.FLOAT32, count=1),
        PointField(name="normal_z", offset=20, datatype=PointField.FLOAT32, count=1),
    ]
    return create_cloud(header, fields, points)


def _get_struct_fmt(
    is_bigendian: bool, fields: List[PointField], field_names: List[str] = None
) -> str:
    """
    Creates struct format string used to create a new python struct object.

    :param is_bigendian: True if native byte order is big-endian
    :param fields: List of pointfields (e.g. x,y,z, rgb etc.)
    :param field_names: Optional list of field names
    """

    fmt = ">" if is_bigendian else "<"
    offset = 0
    for field in (
        f
        for f in sorted(fields, key=lambda f: f.offset)
        if field_names is None or f.name in field_names
    ):
        if offset < field.offset:
            fmt += "x" * (field.offset - offset)
            offset = field.offset
        if field.datatype not in _DATATYPES:
            print("Skipping unknown PointField datatype [%d]" % field.datatype)
        else:
            datatype_fmt, datatype_length = _DATATYPES[field.datatype]
            if field.name == "rgb":
                # For some reason the colour field is having a wrong data type. It should be unsigned int 32 bits. This is a hard-coded fix
                datatype_fmt, datatype_length = _DATATYPES[PointField.UINT32]
            fmt += field.count * datatype_fmt
            offset += field.count * datatype_length

    return fmt


def msg_to_array(msg: PointCloud2) -> np.ndarray:
    """
    Converts data in PointCloud2 msg to numpy array

    :param msg: PointCloud2 msg
    :returns: Numpy array of point data
    """

    all_fields = [field.name for field in msg.fields]
    print("all fields: ", all_fields)
    point_data_generator = read_points(msg, field_names=all_fields, skip_nans=True)

    size = msg.width * msg.height
    dims = len(msg.fields)

    points = np.zeros((size, dims))
    colors = np.zeros((size, 1), dtype=np.int32)  # Seems to be unused??

    i = 0
    fields_order = [i for i in FIELDS_ORDER if i in all_fields]

    for p in point_data_generator:
        p_dict = dict(zip(fields_order, p))

        p_row = np.array([p_dict[k] for k in fields_order])

        if "rgb" in all_fields:
            colors[i] = p_dict["rgb"]
        points[i, :] = p_row

        i += 1

    return points


def o3d_point_cloud_to_msg(
    o3d_cloud: o3d.geometry.PointCloud, frame_id: str = "world"
) -> PointCloud2:
    """
    Converts open3D pointcloud to PointCloud2 message.

    :param o3d_cloud: Open3D point cloud
    :param frame_id: Cloud frame ID to be used in msg header.
    :returns: PointCloud2 message
    """
    empty_header = Header()

    points = np.asarray(o3d_cloud.points)
    size = len(points)

    if o3d_cloud.has_normals():
        normals = np.asarray(o3d_cloud.normals)
    if o3d_cloud.has_colors():
        colors = np.asarray(o3d_cloud.colors)

        def RGBTo32bitInt(r, g, b):
            return np.uint32(int("%02x%02x%02x" % (int(r), int(g), int(b)), 16))

        uint32_colors = np.zeros((size, 1), dtype=np.uint32)

        for i in range(size):
            r = colors[i, 0] * 255.0
            g = colors[i, 1] * 255.0
            b = colors[i, 2] * 255.0

            uint32_colors[i] = RGBTo32bitInt(r, g, b)

    msg = None

    if o3d_cloud.has_normals() and o3d_cloud.has_colors():
        data = np.hstack([points, normals, uint32_colors])
        msg = create_cloud_point_color_normal(empty_header, data)
    elif o3d_cloud.has_normals():
        data = np.hstack([points, normals])
        msg = create_cloud_point_normal(empty_header, data)
    elif o3d_cloud.has_colors():
        data = np.hstack([points, uint32_colors])
        msg = create_cloud_xyz32color(empty_header, data)
    else:
        data = points
        msg = create_cloud_xyz32(empty_header, data)

    msg.header.frame_id = frame_id
    return msg


def array_to_msg(
    data: List[np.ndarray],
    frame_id="world",
    pc_type: PointCloudType = PointCloudType.XYZ,
) -> PointCloud2:
    """
    Converts numpy array of xyz points to PointCloud2 msg.

    :param data: List of iterables, where each iterable corresponds to a point with the data in the same order as the fields (e.g. one point could be [x,y,z,n_x, n_y, n_z, rgb]).
    :param frame_id: Frame ID of point cloud
    :param pc_type: Type defining data present in cloud
    :returns: PointCloud2 message
    """

    empty_header = Header()
    msg = None
    if pc_type == PointCloudType.XYZ:
        msg = create_cloud_xyz32(empty_header, data)
    elif pc_type == PointCloudType.XYZ_NORMAL:
        msg = create_cloud_point_normal(empty_header, data)
    elif pc_type == PointCloudType.XYZ_RGB:
        msg = create_cloud_xyz32color(empty_header, data)
    elif pc_type == PointCloudType.XYZ_NORMAL_RGB:
        msg = create_cloud_point_color_normal(empty_header, data)
    elif pc_type == PointCloudType.XYZ_NORMAL_CURV_RGB:
        msg = create_cloud_point_normal_color_curvature(empty_header, data)

    msg.header.frame_id = frame_id
    return msg


def msg_to_o3d_point_cloud(
    msg: PointCloud2,
    indices: List[int] = None,
    normalise: bool = True,
    vis: bool = False,
) -> o3d.geometry.PointCloud:
    """
    Converts PointCloud2 to open3D point cloud and samples selected indices if specified. Optionally visualises the cloud.
    If colour is present in data, it assumes that RGB is represented by a single 32-bit int.

    NB This function replaces pydvil.utils.pointcloud2_to_o3d() with the same signature.

    :param msg: PointCloud2 message
    :param indices: Point cloud indices to retain. If None, use all.
    :param normalise: If True (default), normals are normalised.
    :param vis: If True, visualise the o3d point cloud.
    :returns: Open3D point cloud.
    """

    data = msg_to_array(msg)

    def toRGBReal(uint_rgb):
        blue = uint_rgb & 255
        green = (uint_rgb >> 8) & 255
        red = (uint_rgb >> 16) & 255
        return np.array([float(red) / 255.0, float(green) / 255.0, float(blue) / 255.0])

    if data.shape[0] > 0:
        pcd = o3d.geometry.PointCloud()
        xyz_points = data[:, :3]
        nan_rows = np.isnan(xyz_points).any(axis=1)
        pcd.points = o3d.utility.Vector3dVector(xyz_points)

        # if msg.fields.normal_x and msg.fields.normal_y and msg.fields.normal_z
        msg_fields = [msg.fields[i].name for i in range(len(msg.fields))]

        if (
            "normal_x" in msg_fields
            and "normal_y" in msg_fields
            and "normal_z" in msg_fields
        ):
            pcd.normals = o3d.utility.Vector3dVector(data[:, 3:6])
            if "rgb" in msg_fields:
                colors = np.zeros((data.shape[0], 3))
                for i in range(data.shape[0]):
                    color = int(data[i, 6])
                    colors[i] = toRGBReal(color)
                pcd.colors = o3d.utility.Vector3dVector(colors)
            if normalise:
                pcd = pcd.normalize_normals()
        elif "rgb" in msg_fields:
            colors = np.zeros((data.shape[0], 3))
            for i in range(data.shape[0]):
                color = int(data[i, 3])
                colors[i] = toRGBReal(color)
            pcd.colors = o3d.utility.Vector3dVector(colors)

        if indices is not None:
            # remove indices corresponding to nan_rows
            indices = np.setdiff1d(np.asarray(indices), np.where(nan_rows))
            pcd_final = pcd.select_by_index(indices)
        else:
            # NaNs will need to be removed before calling segmentation functions
            pcd_final = pcd

        if vis:
            o3d.visualization.draw_geometries([pcd_final.remove_non_finite_points()])

        return pcd_final
    else:
        return None


def point_to_msg(p: List[int] | np.ndarray) -> Point:
    """
    Converts iterable point (xyz) to Point msg.

    :param p: 3x1 iterable representing x,y,z point.
    """

    return Point(x=p[0], y=p[1], z=p[2])


def quaternion_to_msg(q: List[int] | np.ndarray) -> Quaternion:
    """
    Converts iterable quaternion to Quaternion msg.

    :param q: 3x 1 iterable representing quaternion
    """
    return Quaternion(x=q[0], y=q[1], z=q[2], w=q[3])


def position_quaternion_to_msg(
    p: List[int] | np.ndarray, q: List[int] | np.ndarray
) -> Pose:
    """
    Creates a Pose msg from a given xyz location 'p' and quaternion 'q'.

    :param p: 3x1 Iterable representing x,y,z point
    :param q: 3x1 Iterable representing quaternion
    :returns: Pose message
    """
    return Pose(position=point_to_msg(p), orientation=quaternion_to_msg(q))


# Bloc to upright a pointcloud.
def find_rotation_to_xy_plane(plane: np.ndarray) -> np.ndarray:
    """
    Computes the rotation matrix from a given plane (ie, its coefficients) to the XY plane (Normal along Z)

    Args:
        plane: vector4 plane coefficents (a, b, c, d)

    Returns:
        np.ndarray[3,3] rotation matrix to transform input plane to XY plane
    """
    a, b, c, _ = plane
    cos_phi = c / math.sqrt(a * a + b * b + c * c)
    sin_phi = math.sqrt((a * a + b * b) / (a * a + b * b + c * c))
    u1 = b / math.sqrt(a * a + b * b)
    u2 = -a / math.sqrt(a * a + b * b)
    out = np.array(
        [
            [cos_phi + u1 * u1 * (1 - cos_phi), u1 * u2 * (1 - cos_phi), u2 * sin_phi],
            [u1 * u2 * (1 - cos_phi), cos_phi + u2 * u2 * (1 - cos_phi), -u1 * sin_phi],
            [-u2 * sin_phi, u1 * sin_phi, cos_phi],
        ]
    )
    return out


def upright_cloud(scene_cloud, plane_coeffs, flip):
    """
    Transforms a cloud with a plance so that it is oriented so that the plane normal points to Z.
    Args:
        scene_cloud: Pointcloud with a planar surface to transform
        plane_coeffs: Coefficients of the plane on the scene
        flip: flag to invert orientation of the pointcloud
    Returns:
        scene_cloud oriented so that the plane normal is aligned with Z and origin on the centre of the plane.
    """
    scene_upright = copy.deepcopy(scene_cloud)
    scene_upright.translate((0, 0, plane_coeffs[3] / plane_coeffs[2]))
    R = find_rotation_to_xy_plane(plane_coeffs)
    scene_upright.rotate(R, center=(0, 0, 0))
    if flip:
        flip = scene_upright.get_rotation_matrix_from_xyz((np.pi, 0, 0))
        scene_upright.rotate(flip, center=(0, 0, 0))
    return scene_upright



#!/usr/bin/env python

import os
import time
import math
import numpy as np
import pybullet as p

from grip.io import get_data_path
from grip.robot.bullet_robot import BulletRobot
from grip.math.quaternion import position_quaternion2matrix, matrix2position_quaternion
from .end_effector_registry import EndEffectorRegistry, EndEffectorType


base_dir = get_data_path()
PANDA_WITH_ROBOTIQ_2F_85_GRIPPER_URDF = os.path.join(
    base_dir, "urdf/franka_panda/panda_with_robtiq_2f_85_gripper.urdf"
)
PANDA_WITH_ROBOTIQ_2F_140_GRIPPER_URDF = os.path.join(
    base_dir, "urdf/franka_panda/panda_with_robotiq_2f_140_gripper.urdf"
)


class BulletRobotiq2FGripper(BulletRobot, EndEffectorType):
    # Gripper type name for real gripper class (in arm farm layer)
    HW_INTERFACE = "robot_interface.gripper.robotiq_2f_gripper:Robotiq2fGripper"

    def __init__(self, **kwargs):
        kwargs["urdf_file"] = kwargs.get("urdf_file", None)
        kwargs["ee_index"] = kwargs.get("ee_index", -1)
        kwargs["tip_link"] = kwargs.get("tip_link", "tool0")

        print("Tip link: ", kwargs["tip_link"])

        kwargs["joint_names"] = kwargs.get(
            "joint_names",
            [
                "finger_joint",
                "left_inner_finger_joint",
                "left_inner_knuckle_joint",
                "right_outer_knuckle_joint",
                "right_inner_finger_joint",
                "right_inner_knuckle_joint",
            ],
        )

        kwargs["has_kinematics"] = False

        # kwargs["max_force"] = kwargs.get("max_force", 15)

        super().__init__(**kwargs)

        self.T_ee2ik = None
        self.mimic_parent_joint_name = "finger_joint"
        self.ref_joint_name = "finger_joint"
        self.ref_joint_id = self.joint_dict[self.ref_joint_name]["idx"]
        min_angle = self.joint_dict[self.ref_joint_name]["lower_limit"]
        self.max_w = self.angle2width(min_angle)
        max_angle = self.joint_dict[self.ref_joint_name]["upper_limit"]
        self.min_w = self.angle2width(max_angle)

        self.close_angles = [0.700, 0.725, -0.725, -0.725, 0.725, -0.725]
        self.open_angles = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    @property
    def true_ee_pose(self):
        lft_pos, _ = self.get_link_pose(self.link_dict["left_inner_finger_tip"])
        rft_pos, _ = self.get_link_pose(self.link_dict["right_inner_finger_tip"])
        ee_payload_pos = (lft_pos + rft_pos) / 2

        self.get_link_id()
        _, ee_payload_quat = self.get_link_pose(self.tip_link)
        return ee_payload_pos, ee_payload_quat

    @property
    def ee_height(self):
        return self.true_ee_pose[0][2]

    def angle2width(self, angle):
        """
        conversion utility function for robotiq-2f-85
        """
        width = 0.1143 * math.sin(0.715 - angle) + 0.01
        return width

    def width2angle(self, width):
        """
        conversion utility function for robotiq-2f-85
        """
        angle = 0.715 - math.asin((width - 0.01) / 0.1143)
        return angle

    @property
    def width(self):
        """
        Current gripper opening width
        """
        joint_state = p.getJointState(
            self.id, self.ref_joint_id, physicsClientId=self.phys_id
        )
        width = self.angle2width(joint_state[0])
        return width

    def move_to_width(self, width):
        """
        move gripper to the target width

        :param float width: the target width to reach
        """
        target_angle = self.width2angle(width)
        p.setJointMotorControl2(
            self.id,
            self.ref_joint_id,
            p.POSITION_CONTROL,
            targetPosition=target_angle,
            force=self.joint_dict[self.ref_joint_name]["max_force"],
            maxVelocity=self.joint_dict[self.ref_joint_name]["max_vel"],
            physicsClientId=self.phys_id,
        )
        for _ in range(50):
            p.stepSimulation(physicsClientId=self.phys_id)
            time.sleep(1 / 240.0)

    def ee2ik_frame_pose(self, pos, quat, update_T=False):
        """
        conversion function when gripper was fully opened
        """
        if (self.T_ee2ik is None) or update_T:
            ik_frame_pose = position_quaternion2matrix(self.ee_pose[0], self.ee_pose[1])
            ee_frame_pose = position_quaternion2matrix(
                self.true_ee_pose[0], self.true_ee_pose[1]
            )
            self.T_ee2ik = np.matmul(np.linalg.inv(ee_frame_pose), ik_frame_pose)
        pose = position_quaternion2matrix(pos, quat)
        transformed_pose = np.matmul(pose, self.T_ee2ik)
        transformed_pos, transformed_quat = matrix2position_quaternion(transformed_pose)
        return transformed_pos, transformed_quat

    def apply_gripper_delta_action(self, action):
        """
        control gripper with a delta finger distance action.

        :param float action: delta joint angle action for all joints of the phase 4 gripper
        """
        curr_width = self.width
        target_width = min(self.max_w, max(self.min_w, curr_width + action))
        self.move_to_width(target_width)

    def control_fingers(self, mode="close"):
        """
        Opening/closing finger

        :param str mode: 'open' or 'close' mode
        """

        if mode == "close":
            angles = self.close_angles
        else:
            angles = self.open_angles
        for _ in range(5):
            self.set_angles(angles)
            p.stepSimulation(physicsClientId=self.phys_id)
            time.sleep(1 / 240.0)

    def validate_grasp(self, tol: float = 1e-3):
        diff = self.width - 0.012

        if diff < tol:
            has_object = False
        else:
            has_object = True

        return has_object

    def is_ready(self) -> bool:
        """Simulated gripper is always ready

        Returns:
            (bool): whether or not this gripper is ready (in sim it always is)
        """
        return True


class BulletRobotiq2F85Gripper(BulletRobotiq2FGripper, EndEffectorType):
    HW_INTERFACE = "robot_interface.gripper.robotiq_2f_gripper:Robotiq2f85Gripper"

    def __init__(self, **kwargs):
        kwargs["urdf_file"] = kwargs.get(
            "urdf_file", PANDA_WITH_ROBOTIQ_2F_85_GRIPPER_URDF
        )

        super().__init__(**kwargs)

        self.mimic_parent_joint_name = "finger_joint"
        self.mimic_children = {
            "right_outer_knuckle_joint": 1,
            "left_inner_knuckle_joint": 1,
            "right_inner_knuckle_joint": 1,
            "left_inner_finger_joint": -1,
            "right_inner_finger_joint": -1,
        }
        self.setup_mimic_joints(self.mimic_parent_joint_name, self.mimic_children)

        self.close_angles = [0.700, -0.725, 0.725, 0.725, -0.725, 0.725]
        self.open_angles = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.physical_links = [8, 9, 10, 11, 12, 14, 15, 16, 17, 18, 20]
        self.home_positions = self.open_angles
        self.reset_angles(self.home_positions)


@EndEffectorRegistry.register("robotiq_2f140")
class BulletRobotiq2F140Gripper(BulletRobotiq2FGripper, EndEffectorType):
    HW_INTERFACE = "robot_interface.gripper.robotiq_2f_gripper:Robotiq2f85Gripper"

    def __init__(self, **kwargs):
        kwargs["urdf_file"] = kwargs.get(
            "urdf_file", PANDA_WITH_ROBOTIQ_2F_140_GRIPPER_URDF
        )

        super().__init__(**kwargs)

        self.mimic_children = {
            "left_inner_finger_joint": 1,
            "left_inner_knuckle_joint": -1,
            "right_outer_knuckle_joint": -1,
            "right_inner_finger_joint": 1,
            "right_inner_knuckle_joint": -1,
        }
        # self.setup_mimic_joints(self.mimic_parent_joint_name, self.mimic_children)

        self.close_angles = [0.700, 0.725, -0.725, -0.725, 0.725, -0.725]
        self.open_angles = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.physical_links = [8, 9, 10, 11, 12, 14, 15, 16, 17, 18, 20]
        self.home_positions = self.open_angles
        self.reset_angles(self.home_positions)



import numpy as np

import pybullet as p
from grip.robot import BulletObject
from grip.io import log


class ArticulatedObject(BulletObject):
    """
    The ArticulatedObject represents an interface to a Bullet object with multiple movable links.
    An object of this class exposes several utilities for configuring the object's joints of the movable links.

    Args:
        category_name (str): name of the object category.
        instance_id (int): id of the object instance in its category.
        movable_link (dict): a dictionary of the movable link info loaded from the object_meta_info.json in the object dataset. Each key is a movable link name and the value is a dict containing {"link_id", "link_name", "joint_type"}.

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._category_name = kwargs.get("category_name", None)
        self._instance_id = kwargs.get("instance_id", None)
        self._movable_link = kwargs.get("movable_link", None)

        self.joint_states = dict()
        self.links = dict()
        for joint_id in range(p.getNumJoints(self.id, self._phys_id)):
            joint_info = p.getJointInfo(self.id, joint_id, self._phys_id)

            link_name = joint_info[12].decode("gb2312")
            self.links[link_name] = joint_id
            if link_name not in self._movable_link:
                continue
            self.joint_states[joint_id] = {
                "min_val": joint_info[8],
                "max_val": joint_info[9],
                "init_val": p.getJointState(self.id, joint_id, self._phys_id)[0],
            }

    def set_joint(self, joint_id: int, value: float) -> None:
        """
        Sets the joint value of the specified joint.
        If the target state is out of the joint's movable range, this function will log warning.

        Args:
            joint_id (int): id of the selected joint.
            value (float): the target joint value

        """
        if joint_id not in self.joint_states.keys():
            log.warning("Joint {} does not exist.".format(joint_id))

        elif (
            value < self.joint_states[joint_id]["min_val"]
            or value > self.joint_states[joint_id]["max_val"]
        ):
            log.warning("Specified value out of joint operation range.")

        else:
            p.resetJointState(self.id, joint_id, value, self._phys_id)

    def set_joint_random(self, joint_id: int) -> None:
        """
        Sets the joint value of the specified joint to a random value within its range

        Args:
            joint_id (int): id of the selected joint.

        """
        if joint_id not in self.joint_states.keys():
            log.warning("Joint {} does not exist.".format(joint_id))

        else:
            value = np.random.uniform(
                self.joint_states[joint_id]["min_val"],
                self.joint_states[joint_id]["max_val"],
            )
            p.resetJointState(self.id, joint_id, value, self._phys_id)

    def reset(self) -> None:
        """
        Resets all movable joints to their initial values.

        """
        super().reset()
        for joint_id in self.joint_states.keys():
            self.set_joint(joint_id, self.joint_states[joint_id]["init_val"])

    def check_joint_boundary(self, joint_id: int) -> bool:
        """
        Checks if the selected joint has reached its boundary.

        Args:
            joint_id (int): id of the selected joint.

        """
        reach_boundary = False
        self.joint_states[joint_id]["cur_val"] = p.getJointState(
            self.id, joint_id, self._phys_id
        )[0]
        joint_state = self.joint_states[joint_id]
        joint_info = p.getJointInfo(self.id, joint_id, self._phys_id)
        threshold = 0.1 if joint_info[2] == p.JOINT_REVOLUTE else 0.05
        threshold = min(
            threshold, (joint_state["max_val"] - joint_state["min_val"]) / 3.5
        )

        if abs(joint_state["cur_val"] - joint_state["init_val"]) > threshold and (
            joint_state["cur_val"] < joint_state["min_val"] + threshold
            or joint_state["cur_val"] > joint_state["max_val"] - threshold
        ):
            reach_boundary = True

        return reach_boundary



from .logging_utils import *
from .tools import *
from .timer import *

try:
    from grip_data import get_data_path
except:
    from grip_assets import get_data_path


__all__ = [
    "Logger",
    "log",
    "set_log_level",
    "get_package_path",
    "get_data_path",
    "crawl",
    "file_list",
    "get_file_path",
    "pushd",
    "read_file_as_str",
    "Timer",
]



#!/usr/bin/env python3


import os
import numpy as np

try:
    import rclpy

except ImportError:
    raise RuntimeError("Unable to run example without ros2 installed.")

import pybullet_data as pd
from grip.ros import ROSRobotArm, ROSCamera
import grip
from grip.ros import xacro_to_urdf_path


class ROSArmExample(rclpy.node.Node):
    """This node example will simulate a robot arm with a wrist camera."""

    def __init__(self, camera_fps: int = 10, joint_pub_rate: int = 150):
        super().__init__("robot_arm_example_node")

        # Creates bullet world with GUI
        self.world = grip.robot.BulletWorld(phys_opt="gui")

        # Creates ground plane object
        self.plane = grip.robot.BulletObject(
            urdf_file=os.path.join(pd.getDataPath(), "plane.urdf"), position=[0, 0, 0]
        )

        self.plane.set_colour([0.0, 0.7, 0.7, 1.0])
        self.duck = grip.robot.BulletObject(
            urdf_file=os.path.join(pd.getDataPath(), "duck_vhacd.urdf"),
            position=[0.7, 0.0, 0.0],
            scale=2.0,
            fixed=False,
        )

        urdf_path = os.path.join(pd.getDataPath(), "franka_panda/panda.urdf")

        # urdf_path = xacro_to_urdf_path(
        #     "franka_description",
        #     "panda_arm.urdf.xacro",
        #     "/tmp/panda_arm.urdf",
        #     pkg_subfolder="robots",
        #     xacro_args="hand:=true",
        # )

        # Creates robot arm
        self.arm = ROSRobotArm(
            parent_node=self,
            urdf_file=urdf_path,
            phys_id=self.world.id,
            ee_mount_link="panda_link8",
            tip_link="panda_grasptarget",  # "panda_hand_tcp",
            gripper_type="panda_hand",
            sliders=True,
        )

        print("Links: ", self.arm.get_link_names())

        # Creates ROSCamera object
        self.camera = ROSCamera(
            name="hand_camera",
            cid=self.world.id,
            position=np.array([0, 0, 1.5]),
            orientation=grip.math.rpy2quaternion([0.0, np.pi, 0.0]),
            parent_node=self,
            anchor_robot=self.arm,
            anchor_link="panda_grasptarget",  # "panda_hand_tcp",
            base_frame_id="panda_link0",
            sim=self.arm.sim,
        )

        if self.arm.sim:
            # If we are in sim, then we create timer publishers for camera and joint states
            # This is not needed if data is to be consumed synchrnously.
            # However, it is needed if one wants to simulate asynchronous publisers just like the real robot and cameras would be behave.

            # Start camera timer publisher
            self.camera.setup_timer_publisher(fps=camera_fps)
            self.arm.joint_state_pub.setup_timer_publisher(rate=joint_pub_rate)

            self.timer = self.create_timer(1 / camera_fps, self.on_update)

            grip.robot.add_text([0, 0, 1], "mode: robot is simulated")
        else:
            grip.robot.add_text([0, 0, 1], "mode: robot is mirroring real robot state")

            self.send_goal_button = grip.robot.gui.BulletButton(
                self.world.id, "Execute Joint Goal"
            )

            def on_update():
                joint_goal = self.arm.joint_sliders.read()

                grip.robot.add_debug_frame(*self.arm.forward_kinematics(joint_goal))

                if self.send_goal_button.was_pressed():
                    path = grip.motion.Path(
                        points=joint_goal.reshape(1, -1),
                        joint_names=self.arm.joint_names,
                        times=np.array([1.0]),
                    )

                    self.arm.follow_joint_path(path)

            self.timer = self.create_timer(1 / camera_fps, on_update)

    def on_update(self):
        arm_positions = self.arm.joint_sliders.read()
        gripper_positions = self.arm.gripper.joint_sliders.read()

        if self.arm.joint_sliders.has_changed():
            self.arm.set_angles(arm_positions)

        if self.arm.gripper.joint_sliders.has_changed():
            self.arm.gripper.set_angles(gripper_positions)


def main(args=None):
    rclpy.init(args=args)

    ros_arm_example = ROSArmExample()

    rclpy.spin(ros_arm_example)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    ros_arm_example.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()



# Copyright (c) 2008, Willow Garage, Inc.
# All rights reserved.
#
# Software License Agreement (BSD License 2.0)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of the Willow Garage nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Optional
from launch import LaunchDescription

import launch_ros.actions
import launch_ros.descriptions


def depth_proc_launch(
    camera_name: Optional[str] = "CameraSensor",
    camera_info_topic: Optional[str] = "/hand_camera/color/undistorted/camera_info",
    colour_image_topic: Optional[str] = "/hand_camera/color/undistorted/image_rect",
    depth_image_topic: Optional[
        str
    ] = "/hand_camera/depth_registered/undistorted/image_rect",
) -> LaunchDescription:
    """Creates the launch description for a depth processing node, which converts rgb, depth and camera info to a point cloud with xyzrgb fields.

    Args:
        camera_name (Optional[str], optional): camera name which is used as prefix for the topics this depth proccessing node will subscribe to. Defaults to "CameraSensor".

    Returns:
        LaunchDescription: the launch description for a depth processing node
    """

    return LaunchDescription(
        [
            # launch plugin through rclcpp_components container
            launch_ros.actions.ComposableNodeContainer(
                name="container",
                namespace="",
                package="rclcpp_components",
                executable="component_container",
                composable_node_descriptions=[
                    # Driver itself
                    launch_ros.descriptions.ComposableNode(
                        package="depth_image_proc",
                        plugin="depth_image_proc::PointCloudXyzrgbNode",
                        name="point_cloud_xyzrgb_node",
                        remappings=[
                            ("rgb/camera_info", camera_info_topic),
                            (
                                "rgb/image_rect_color",
                                colour_image_topic,
                            ),
                            (
                                "depth_registered/image_rect",
                                depth_image_topic,
                            ),
                            ("points", f"/{camera_name}/depth_registered/points"),
                        ],
                    ),
                ],
                output="screen",
            ),
        ]
    )


def generate_launch_description(camera_name="CameraSensor"):
    return depth_proc_launch(camera_name)



import pybullet as p
from .entity import Entity
from .types import BulletContactInfo


def get_collision_info(entity_a: Entity, entity_b: Entity, max_distance: float = 0.0):
    """Returns collision information from pair of entities

    Args:
        entity_a (Entity): the first entity
        entity_b (Entity): the second entity
        max_ditance (float, optional): If the distance between objects is greater than this maximum distance, no points are be returned.
    Returns:
        (List[BulledContactInfo]): a list of contacts if any exists
    """

    p.performCollisionDetection()
    contacts = []

    links_a = entity_a.get_link_ids()
    links_b = entity_b.get_link_ids()

    for l_a in links_a:
        for l_b in links_b:
            contacts += p.getClosestPoints(
                entity_a.id,
                entity_b.id,
                linkIndexA=l_a,
                linkIndexB=l_b,
                distance=max_distance,
                physicsClientId=entity_a.phys_id,
            )

    return [BulletContactInfo(*c) for c in contacts]


def is_colliding(entity_a: Entity, entity_b: Entity, max_distance: float = 0.0):
    """Returns collision information from pair of entities

    Args:
        entity_a (Entity): the first entity
        entity_b (Entity): the second entity
        max_ditance (float, optional): If the distance between objects is greater than this maximum distance, no points are be returned.
    Returns:
        (bool): whether or not collision contacts exist
    """

    return len(get_collision_info(entity_a, entity_b, max_distance)) != 0



#!/usr/bin/env python3
import os
import time
import grip
import pybullet_data as pd

if __name__ == "__main__":
    world = grip.robot.BulletWorld(phys_opt="gui")

    urdf_path = os.path.join(pd.getDataPath(), "franka_panda/panda.urdf")

    robot = grip.robot.BulletRobot(urdf_file=urdf_path, sliders=True, phys_id=world.id)

    plane = grip.robot.BulletObject(
        urdf_file=os.path.join(pd.getDataPath(), "plane.urdf"), position=[0, 0, 0]
    )

    while world.is_connected():
        positions = robot.joint_sliders.read()

        robot.set_angles(positions)

        time.sleep(0.1)
